"""Base extractor IAction implementation.

The BaseExtractor class provides an abstract extractor implementation.
 
"""

from GangaRobot.Framework.Action import IAction
from GangaRobot.Framework import Utility
from GangaRobot.Lib.Base.Extract import Node
from GangaCore.Utility.logging import getLogger
from GangaCore.GPI import *

logger = getLogger()


class BaseExtractor(IAction):
    
    """Base extractor IAction implementation.
    
    An abstract action implementation providing a basis for concrete run data
    extractor implementations. It contains a chain of extractors which are
    processed in the execute() method by calling handlerunnode() once, and
    handlejobnode() for each job associated with the runid, on each extractor
    in turn.
    
    Implementing a sub-class involves doing any of the following:
    - Override the __init__() method to add additional extractors to the chain
    attribute.
    - Override the handlerunnode() method to add extracted data to the run node.
    - Override the handlejobnode() method to add extracted data to the job node
    for the given job.
    
    The extracted run node has the structure:
    <run>
        <job/>
        <job/>
        ...
    </run>
    
    """
    
    def __init__(self):
        """Create a new base extractor.
        
        The chain attribute, a list of extractors, is initialised to [self].
        
        Sub-classes can override this constructor to set the chain attribute.
        e.g.
        self.chain = [CoreExtractor(), self]

        """
        self.chain = [self]

    def execute(self, runid):
        """Invoke each of the extractors in the chain.
        
        Keyword arguments:
        runid -- A UTC ID string which identifies the run.

        An empty run node is created as an instance of Extract.Node and passed
        to the handlerunnode() method of each extractor in the chain.
        
        An empty job node is created, as a subnode of the run node, for each job
        in the jobtree directory named by the runid, e.g. /2007-06-25_09.18.46,
        and passed to the handlejobnode() method of each extractor in the chain.
        
        The run node is saved as XML to the configurable BaseExtractor_XmlFile,
        replacing ${runid} with the current run id.
        e.g. ~/gangadir/robot/extract/${runid}.xml
        
        """
        logger.info("Extracting data for run '%s'.", runid)
        runnode = Node('run')
        for extractor in self.chain:
            extractor.handlerunnode(runnode, runid)
        path = Utility.jobtreepath(runid)
        ids = jobtree.listjobs(path)
        ids.sort()
        for id in ids:
            jobnode = Node('job')
            job = jobs(id)
            for extractor in self.chain:
                extractor.handlejobnode(jobnode, job)
            runnode.nodes.append(jobnode)
        self._saveextract(runnode, runid)
        logger.info('Extract data saved.')

    def _saveextract(self, runnode, runid):
        filename = Utility.expand(self.getoption('BaseExtractor_XmlFile'), runid = runid)
        content = runnode.toprettyxml()
        Utility.writefile(filename, content)

    def handlerunnode(self, runnode, runid):
        """Empty default implementation.
        
        Keyword arguments:
        runnode -- An Extract.Node for the run.
        runid -- A UTC ID string which identifies the run.
        
        Sub-classes can override this method to add extracted data to the run
        node, possibly using the Ganga GPI.
        e.g.
        def handlerunnode(self, runnode, runid):
            corenode = runnode.addnode('core')
            corenode.addnode('id', runid)

        This method is called in the context of a Ganga session.
        
        """
        pass

    def handlejobnode(self, jobnode, job):
        """Empty default implementation.
        
        Keyword arguments:
        jobnode -- An Extract.Node for the given job.
        job -- A Ganga Job associated with the run.
        
        Sub-classes can override this method to add extracted data to the job
        node, possibly using the Ganga GPI.
        e.g.
        def handlejobnode(self, jobnode, job):
            corenode = jobnode.addnode('core')
            corenode.addnode('id', job.id)

        This method is called in the context of a Ganga session.
        
        """
        pass

    




###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: Schema.py,v 1.3 2009-05-20 13:40:22 moscicki Exp $
##########################################################################

import GangaCore.Utility.logging

import copy

from GangaCore.Utility.logic import implies

import GangaCore.Utility.Config

from GangaCore.Core.exceptions import GangaAttributeError, TypeMismatchError, SchemaError, GangaException, GangaTypeError

from GangaCore.Utility.Plugin import allPlugins

import types

from GangaCore.GPIDev.TypeCheck import _valueTypeAllowed

from inspect import isclass

from GangaCore.Utility.Config import Config

logger = GangaCore.Utility.logging.getLogger()

# Dictionary for storing data from the Config system which takes a while to lookup
_found_components = {}
_found_configs = {}
_found_attrs = {}
_stored_defaults = {}
_stored_options = {}
_stored_configs = {}

#
# Ganga Public Interface Schema
#

# Version of the schema.


class Version(object):

    __slots__ = ('major', 'minor')

    def __init__(self, major, minor):
        self.major = major
        self.minor = minor

    def __eq__(self, v):
        return self.major == v.major and self.minor == v.minor

    def __ne__(self, v):
        return not self == v

    def isCompatible(self, v):
        return v.major == self.major and v.minor <= self.minor


def defaultConfigSectionName(name):
    global _stored_defaults
    try:
        return _stored_defaults[name]
    except KeyError:
       _stored_defaults[name] = 'defaults_' + name  # _Properties
       return _stored_defaults[name]

# Schema defines the logical model of the Ganga Public Interface (GPI)
# for  jobs and  pluggable  job components  such  as applications  and
# submission backends.

# Schema contains the data description (properties) which is the basis
# of the efficient management of the persistent data in a generic way.
# Any Ganga component is  a container of properties (a property
# may be a component as well).  Properties may be protected (read-only
# by  the  users).  Persistent  properties  are  stored  with the  job
# whereas transient are not.  This additional information is contained
# in  metaproperties. Large fractions of Ganga GPI may be generated
# automatically based on the metaproperties definition in a schema.

# Each  component   belongs  to   a  certain  category   (for  example
# 'application'). There  may be many different  types of applications,
# each one defined by a separate component.

# Schema  of each  component  provides a  version  information (it  is
# important for persistent storage) and compatibility of the scripts.

# Internal Note:  Schema object must be  additionally initialized with
# the _pluginclass object  which is used as the  source of information
# about name and category of  the schema deepcopy of Schema objects is
# possible  however   the  _pluginclass  objects   are  shared  unless
# overriden explicitly.

class Schema(object):
    # Schema constructor is used by Ganga plugin developers.
    # Ganga will automatically set a reference to the plugin class which corresponds to this schema, hence
    # defining the schema's name and category.
    #
    # datadict: dictionary of properties (schema items) Defaults to '{}'
    # version: the version information

    __slots__ = ('datadict', 'version', '_pluginclass')

    def __init__(self, version, datadict=None):
        self.datadict = datadict or {}
        self.version = version
        self._pluginclass = None

    def __getitem__(self, name):
        try:
            return self.datadict[name]
        except KeyError:
            if self._pluginclass is not None:
                err_str = "Ganga Cannot find: %s in Object: %s" % (name, self.name)
            else:
                err_str = "Ganga Cannot find: %s attribute for None object" % (name,)
            logger.error(err_str)
            raise GangaAttributeError(err_str)

    @property
    def category(self):
        return self._pluginclass._category

    @property
    def name(self):
        return self._pluginclass._name

    def allItemNames(self):
        return list(self.datadict.keys())

    def allItems(self):
        if self.datadict is None: return list(zip())
        return list(zip(list(self.datadict.keys()), list(self.datadict.values())))

    def simpleItems(self):
        return self._filter(SimpleItem)

    def componentItems(self):
        return self._filter(ComponentItem)

    def sharedItems(self):
        return self._filter(SharedItem)

    def hasAttribute(self, name):
        return name in self.datadict

    def getItem(self, name):
        return self.__getitem__(name)

    def hasItem(self, name):
        return name in self.datadict

    def _filter(self, klass):
        if self.datadict is None:
            return []

        r = [(n, c) for (n, c) in self.datadict.items() if issubclass(c.__class__, klass)]
        return r

    def isEqual(self, schema):
        return self.name == schema.name and self.category == schema.category

    # make a schema copy for a derived class, does not copy the pluginclass
    def inherit_copy(self):
        new_dict = {}
        for key, val in self.datadict.items():
            new_dict[key] = copy.deepcopy(val)
        return Schema(copy.deepcopy(self.version), new_dict)

    def createDefaultConfig(self):
        # create a configuration unit for default values of object properties
        # take the defaults from schema defaults
        _self_name = self.name
        config = GangaCore.Utility.Config.makeConfig(defaultConfigSectionName(_self_name), "default attribute values for %s objects" % _self_name)

        for name, item in self.allItems():
            # and not item['sequence']: #FIXME: do we need it or not??
            if not item['protected'] and not item['hidden']:
                if item.hasProperty('typelist'):
                    types = item['typelist']
                    if types == []:
                        types = None
                else:
                    types = None

                if item['sequence']:
                    if not types is None:
                        # bugfix 36398: allow to assign a list in the
                        # configuration
                        types.append('list')
                if isinstance(item['defvalue'], dict):
                    if not types is None:
                        types.append('dict')
                config.addOption(name, item['defvalue'], item['doc'], False, typelist=types)


        def prehook(name, x):
            errmsg = "Cannot set %s=%s in [%s]: " % (name, repr(x), config.name)

            try:
                item = self.getItem(name)
            except Exception as x:
                raise GangaCore.Utility.Config.ConfigError(errmsg + str(x))

            if item.isA(ComponentItem):
                if not isinstance(x, str) and not x is None:
                    raise GangaCore.Utility.Config.ConfigError(errmsg + "only strings and None allowed as a default value of Component Item.")
                try:
                    self._getDefaultValueInternal(name, x, True)
                except Exception as err:
                    logger.info("Unexpected error: %s", err)
                    raise

            if item['protected'] or item['hidden']:
                raise GangaCore.Utility.Config.ConfigError(errmsg + "protected or hidden property")

            return x

        config.attachUserHandler(prehook, None)
        config.attachSessionHandler(prehook, None)


    def getDefaultValue(self, attr, make_copy=True):
        """ Get the default value of a schema item, both simple and component.
        attr, some attribute name
        self, myself a Schema
        make_copy, should I return the default object of a deepcopy of it?
        """
        returnable = self._getDefaultValueInternal(attr)
        if make_copy and returnable is not None:
            from GangaCore.GPIDev.Base.Objects import GangaObject
            if isinstance(returnable, GangaObject):
                return returnable.__class__.getNew(should_init=True)
            else:
                return copy.deepcopy(returnable)
        else:
            return returnable

    def _getDefaultValueInternal(self, attr, val=None, check=False):
        """ Get the default value of a schema item, both simple and component.
        If check is True then val is used instead of default value: this is used to check if the val may be used as a default value (e.g. if it is OK to use it as a value in the config file)
        """
        def_name = defaultConfigSectionName(self.name)

        item = self.getItem(attr)

        stored_option_key = def_name
        stored_attr_key = def_name + ':' + attr

        is_finalized = Config._after_bootstrap

        useDefVal = False

        try:

            # Attempt to get the relevant config section
            if is_finalized:
                if stored_option_key in _stored_configs:
                    config = _stored_configs[stored_option_key]
                else:
                    config = Config.getConfig(def_name)
                    _stored_configs[stored_option_key] = config
            else:
                config = Config.getConfig(def_name)

            if is_finalized and not config.hasModified and stored_attr_key in _found_attrs:
                defvalue = _found_attrs[stored_attr_key]
            else:
                if is_finalized:
                    if stored_option_key in _stored_options and not config.hasModified:
                        eff_ops = _stored_options[stored_option_key]
                    else:
                        eff_ops = config.getEffectiveOptions()
                        _stored_options[stored_option_key] = eff_ops
                else:
                    eff_ops = config.getEffectiveOptions()

                if attr in eff_ops:
                    defvalue = config[attr]
                    from GangaCore.GPIDev.Base.Proxy import isProxy
                    if isProxy(defvalue):
                        raise GangaException("(1)Proxy found where it shouldn't be in the Config: %s" % stored_attr_key)
                    ## Just in case a developer puts the proxied object into the default value!
                    _found_attrs[stored_attr_key] = defvalue
                else:
                    useDefVal = True
        except (KeyError, Config.ConfigError):
            useDefVal = True

        if useDefVal:
            # hidden, protected and sequence values are not represented in config
            defvalue = item['defvalue']
            from GangaCore.GPIDev.Base.Proxy import isProxy
            if isProxy(defvalue):
                raise GangaException("(2)Proxy found where is shouldn't be in the Config" % stored_attr_key)
            ## Just in case a developer puts the proxied object into the default value!
            _found_attrs[stored_attr_key] = defvalue

        # in the checking mode, use the provided value instead
        if check is True:
            defvalue = val

        if isinstance(item, ComponentItem):

            # FIXME: limited support for initializing non-empty sequences (i.e.
            # apps => ['DaVinci','Executable'] is NOT correctly initialized)

            if not item['sequence']:
                if defvalue is None:
                    if not item['load_default']:
                        try:
                            assert(item['optional'])
                        except AssertionError:
                            raise SchemaError("This item '%s' is not a sequence, doesn't have a load_default and is not optional. This is unsupported!" % type(item))
                        return None

                # if a defvalue of a component item is an object (not string) just process it as for SimpleItems (useful for FileItems)
                # otherwise do a lookup via plugin registry

                category = item['category']

                if isinstance(defvalue, str) or defvalue is None:
                    try:
                        config = Config.getConfig(def_name)
                        has_modified = config.hasModified
                    except KeyError:
                        has_modified = False

                    if category not in _found_components or has_modified:
                        _found_components[category] = allPlugins.find(category, defvalue)
                    return _found_components[category]()
                if isclass(defvalue):
                    return defvalue()

        # If needed/requested make a copy of the function elsewhwre
        return defvalue


# Items in schema may be either Components,Simples, Files or BindingItems.
#
# Component, Simple, File are data items and have implicit storage.
#
# Metaproperties of all Items:
#
# transient : never stored on persistent media
# protected : not modifiable via GPI
# hidden    : not visible in the GPI
# comparable: taken into account for equality checks
# sequence:   an item is a sequence (algorithms traversing the component tree dive into sequences as well)
# strict_sequence: if not strict sequence then assignment of a single item will be automatically converted to a 1-element sequence, i.e. obj.x = v => obj.x = [v]
# defvalue  : default value, if item is a sequence the defvalue must be a list
#             for component this must be either None or a string with a name of the component in the same category
# copyable  : if 0 then the property value will not be copied and the default value from schema will be set in the destination
#             implied initialization rule (unless the "copyable" metaproperty is explicitly set)
#             protected == 1 => copyable == 0
#             protected == 0 => copyable == 1
# doc       : a docstring
# checkset  : a bound checkset method, restrict write access at the object level (for example job.status
#             may not be modified directly in backend handlers, instead updateStatus() method should be used)
#
# filter    : a bound method can be used to change the values of an attribute. The attribute will be set to the return value
#
# visitable : if false then all algorithms based on the visitor patter will not accept this item [true]
#             this is needed in certain cases (such as job.master) to avoid infinite recursion (and loops)
#
# summary_print: An bound method name (string). Will be passed an attribute value and a verbosity_level. Should
#                return a (Python parsable) string summarising the state of the value.
#
# summary_sequence_maxlen: An integer longer than which a sequence will be summerised when doing a summary
#                          print. If the value is -1, the sequence will never be summerised.
#
# changable_at_resubmit: if true than the value may be changed at job resubmission, see Job.resubmit()
#
#
# Metaproperties of SimpleItems
#
# typelist  : a list of type names (strings) indicating allowed types of the property (e.g. ["str","int","GangaCore.GPIDev.Lib.File.File.File"]), see: http://twiki.cern.ch/twiki/bin/view/ArdaGrid/GangaTypes
#     please consider not using strings in future, we can support the type objects correctly
#
#
# Metaproperties of ComponentItems:
#
# category : category of the component ('applications','backends',...)
# optional : if true then None may be used as a legal value of the item, [false]
# load_default: if true and defvalue is None then load default plugin, [true]
#
# defvalue is None load_default==true   => load default plugin
# defvalue is None load_default==false  => use None (optional MUST be true, otherwise error)
#
# defvalue is 'x'  => load 'x'
#
# getter : a bound getter method, this implies that the component does not have associated storage and cannot be neither set nor deleted [None]
#          getter implies: transient=1, protected=1, sequence=0, defvalue=None, load_default=0, optional=1, copyable=0
# proxy_get: a bound getter method for proxy decoration, allows to customize completely the creation of the proxy
#
#
# Metaproperties of FileItems
#
# FIXME: under development
# in/out    : direction wrt to the job submission operation
# ref/copy  : reference vs copy semantics for file contents
#

# BindingItems enable to hook arbitrary getter methods and do not have any implicit storage. Therefore
# the BindingItems are always transient, cannot be copied, cannot be
# sequences and have None default value.

class Item(object):
    # default values of common metaproperties
    _metaproperties = {'transient': 0, 'protected': 0, 'hidden': 0, 'comparable': 1, 'sequence': 0, 'defvalue': None, 'copyable': 1,
                        'doc': '', 'visitable': 1, 'checkset': None, 'filter': None, 'strict_sequence': 1, 'summary_print': None,
                        'summary_sequence_maxlen': 5, 'proxy_get': None, 'getter': None, 'changable_at_resubmit': 0, 'preparable': 0,
                        'optional': 0, 'category': 'internal', 'typelist': None, 'load_default': 1}

    __slots__ = ('_meta')

    def __init__(self):
        super(Item, self).__init__()
        self._meta = copy.deepcopy(Item._metaproperties)

    def __construct(self, args):
        self._meta = copy.deepcopy(Item._metaproperties)

    def __getitem__(self, key):
        return self._meta[key]

    def hasProperty(self, key):
        return key in self._meta

    def getProperties(self):
        return self._meta

    def __len__(self):
        return len(self._meta)

    # compare the kind of item:
    # all calls are equivalent:
    # item.isA(SimpleItem)
    def isA(self, _what):
        from GangaCore.GPIDev.Base.Proxy import stripProxy

        what = stripProxy(_what)
        return issubclass(self.__class__, what)


    def _update(self, kwds, forced=None):
        """ Add new metaproperties/override old values. To be used by derived contructors only.
        'forced' is an (optional) dictionary containing all values which cannot be modified by
        the user of the derived contructor.
        """
        if forced:
            # find intersection
            forbidden = [k for k in forced if k in kwds]
            if len(forbidden) > 0:
                raise GangaTypeError('%s received forbidden (forced) keyword arguments %s' % (self.__class__, forbidden))
            self._meta.update(forced)

        self._meta.update(kwds)

        # conditional initial value logic...
        # unless 'copyable' explicitly set (or forced), protected==1 =>
        # copyable==0
        if 'copyable' not in kwds and (not forced or 'copyable' not in forced):
            if self._meta['protected']:
                #logger.debug('applied implicit conditional rule: protected==1 => copyable==0')
                self._meta['copyable'] = 0

    # return a description of the property including a docstring
    def describe(self):

        specdoc = '(' + self._describe() + ')'

        if self['doc']:
            s = "%s. %s" % (self['doc'], specdoc)
        else:
            s = specdoc

        return s

    # return a meta description of the property
    def _describe(self):
        first = ','
        #txt = ' ['
        txt = ''
        for m in ['transient', 'protected', 'comparable', 'optional']:
            if m in self._meta:
                txt += '%s%s' % (first, m)
            # try:
            #    if self[m]:
            #        txt += '%s%s'%(first,m)
            #        first = ','
            # except KeyError:
            #    pass
        #txt += ']'
        txt = " default=" + repr(self['defvalue']) + txt
        if self['sequence']:
            txt = " list," + txt
        return txt

    @staticmethod
    def __check(isAllowedType, name, validTypes, input_val):
        if not isAllowedType:
            #import traceback
            #traceback.print_stack()
            raise TypeMismatchError('Attribute "%s" expects a value of the following types: %s\nfound: "%s" of type: %s' % (name, validTypes, input_val, type(input_val)))

    def _check_type(self, val, name, enableGangaList=True):

        if enableGangaList:
            from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
        else:
            GangaList = list

        # type checking does not make too much sense for Component Items because component items are
        # always checked at the object (_impl) level for category compatibility.

        if self.isA(ComponentItem):
            return

        validTypes = self._meta['typelist']

        # setting typelist explicitly to None results in disabling the type
        # checking completely
        if validTypes is None:
            return

        if self._meta['sequence']:
            from GangaCore.GPIDev.Base.Proxy import isType
            if not isType(self._meta['defvalue'], (list, tuple, GangaList)):
                raise SchemaError('Attribute "%s" defined as a sequence but defvalue is not a list.' % name)

            if not isType(val, (GangaList, tuple, list)):
                raise TypeMismatchError('Attribute "%s" expects a list.' % name)

        if validTypes:
            if self._meta['sequence']:

                for valItem in val:
                    if not valueTypeAllowed(valItem, validTypes):
                        raise TypeMismatchError('List entry %s for %s is invalid. Valid types: %s' % (valItem, name, validTypes))

                return
            else:  # Non-sequence
                if isinstance(self._meta['defvalue'], dict):
                    if isinstance(val, dict):
                        for dKey, dVal in val.items():
                            if not valueTypeAllowed(dKey, validTypes) or not valueTypeAllowed(dVal, validTypes):
                                raise TypeMismatchError('Dictionary entry %s:%s for attribute %s is invalid. Valid types for key/value pairs: %s' % (dKey, dVal, name, validTypes))
                    else:  # new value is not a dict
                        if val == []:
                            return
                        raise TypeMismatchError('Attribute "%s" expects a dictionary. Found: "%s".' % (name, type(val)))
                    return
                else:  # a 'simple' (i.e. non-dictionary) non-sequence value
                    self.__check(valueTypeAllowed(val, validTypes), name, validTypes, val)
                    return

        # typelist is not defined, use the type of the default value
        if self._meta['defvalue'] is None:
            logger.warning('type-checking disabled: type information not provided for %s, contact plugin developer', name)
        else:
            if self._meta['sequence']:
                logger.warning('type-checking is incomplete: type information not provided for a sequence %s, contact plugin developer', name)
            else:

                logger.debug("valType: %s defValueType: %s name: %s" % (type(val), type(self._meta['defvalue']), name))
                self.__check(self.__actualCheck(val, self._meta['defvalue']), name, type(self._meta['defvalue']), val)


    @staticmethod
    def __actualCheck(val, defVal):

        try:
            from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
            knownLists = (list, tuple, GangaList)
        except Exception as err:
            knownLists = (list, tuple)
        from GangaCore.GPIDev.Base.Proxy import isType
        if isType(defVal, knownLists) and isType(val, knownLists):
                return True
        else:
            if type(defVal) == type:
                return isType(val, type)
            else:
                return isType(val, type(defVal))


class ComponentItem(Item):
    # schema user of a ComponentItem cannot change the forced values below
    _forced = {}

    __slots__=list()

    def __init__(self, category, optional=0, load_default=1, **kwds):
        super(ComponentItem, self).__init__()
        kwds['category'] = category
        kwds['optional'] = optional
        kwds['load_default'] = load_default
        #kwds['getter'] = getter
        self._update(kwds, ComponentItem._forced)
        try:
            assert(implies(self['defvalue'] is None and not self['load_default'], self['optional']))
        except AssertionError:
            raise SchemaError("ComponentItem has no defvalue, load_default or requirement to be optional")

        try:
            assert(implies(self['getter'], self['transient'] and self['defvalue'] is None and self['protected'] and not self['sequence'] and not self['copyable']))
        except AssertionError:
            raise SchemaError("There is no getter, transient flag or defvalue and the ComponentItem is protected, not a sequence or not copyable. This is not supported")

    def _describe(self):
        return "'" + self['category'] + "' object," + Item._describe(self)


valueTypeAllowed = lambda val, valTypeList: _valueTypeAllowed(val, valTypeList, logger)

defaultValue='_NOT_A_VALUE_'


class SimpleItem(Item):

    __slots__=list()

    def __init__(self, defvalue, typelist=defaultValue, **kwds):
        super(SimpleItem, self).__init__()
        if typelist == defaultValue:
            if type(defvalue) == dict:
                typelist = []
            else:
                typelist = [type(defvalue)]
        kwds['defvalue'] = defvalue
        kwds['typelist'] = typelist
        self._update(kwds)

    def _describe(self):
        return 'simple property,' + Item._describe(self)


class SharedItem(Item):

    __slots__=list()

    def __init__(self, defvalue, typelist=defaultValue, **kwds):
        super(SharedItem, self).__init__()
        if typelist == defaultValue:
            if type(defvalue) == dict:
                typelist = []
            else:
                typelist = [type(defvalue)]
        kwds['defvalue'] = defvalue
        kwds['typelist'] = typelist
        self._update(kwds)

    def _describe(self):
        return 'shared property,' + Item._describe(self)


# Files are important and common enough to merit a special support for
# defining their metaproperties
class FileItem(ComponentItem):

    __slots__=list()

    def __init__(self, **kwds):
        super(FileItem, self).__init__('files')
        self._update(kwds)

    def _describe(self):
        return "'files' object," + Item._describe(self)


class GangaFileItem(ComponentItem):

    __slots__ = list()

    def __init__(self, **kwds):
        super(GangaFileItem, self).__init__('gangafiles')
        self._update(kwds)

    def _describe(self):
        return "'gangafiles' object," + Item._describe(self)


if __name__ == '__main__':

    # a simple test

    dd = {
        'application': ComponentItem(category='applications'),
        'backend':     ComponentItem(category='backends'),
        'name':        SimpleItem('', comparable=0),
        'workdir':     SimpleItem(defvalue=None, type='string', transient=1, protected=1, comparable=0),
        'status':      SimpleItem(defvalue='new', protected=1, comparable=0),
        'id':           SimpleItem(defvalue=None, type='string', protected=1, comparable=0),
        'inputbox':     FileItem(defvalue=[], sequence=1),
        'outputbox':    FileItem(defvalue=[], sequence=1),
        'overriden_copyable': SimpleItem(defvalue=None, protected=1, copyable=1),
        'plain_copyable': SimpleItem(defvalue=None, copyable=0)
    }

    schema = Schema(Version(1, 0), dd)

    # NOT a public interface: emulate the Ganga Plugin object for test purposes
    # Note that pclass MUST be a new-style class in order to support deepcopy
    class pclass(object):
        _category = 'jobs'
        _name = 'Job'
    schema._pluginclass = pclass
    # end of emulating code
    # allSchemas.add(schema)

    assert(schema.name == 'Job')
    assert(schema.category == 'jobs')

    assert(schema.allItems() == list(dd.items()))

    cc = (schema.componentItems() + schema.simpleItems()).sort()
    cc2 = list(dd.items()).sort()
    assert(cc == cc2)

    for i in schema.allItems():
        assert(schema[i[0]] == schema.getItem(i[0]))

    assert(schema['id'].isA(SimpleItem))
    assert(schema['application'].isA(ComponentItem))
    assert(schema['inputbox'].isA(ComponentItem))
    assert(schema['inputbox'].isA(FileItem))

    assert(schema['id']['protected'])
    assert(not schema['id']['comparable'])
    assert(schema['id']['type'] == 'string')

    logger.info(schema['application']['category'] + ' ' + schema['application']['defvalue'])

    schema2 = copy.deepcopy(schema)

    assert(schema2 is not schema)
    assert(schema.datadict is not schema2.datadict)
    assert(schema._pluginclass is schema2._pluginclass)

    for i in schema.allItems():
        assert(schema.getItem(i[0]) is not schema2.getItem(i[0]))

    # check the implied rules

    assert(schema['overriden_copyable']['copyable'] == 1)
    assert(schema['plain_copyable']['copyable'] == 0)
    assert(schema['id']['copyable'] == 0)
    assert(schema['application']['copyable'] == 1)

    logger.info('Schema tested OK.')


#
#
# $Log: not supported by cvs2svn $
# Revision 1.2  2008/09/09 14:37:16  moscicki
# bugfix #40220: Ensure that default values satisfy the declared types in the schema
#
# factored out type checking into schema module, fixed a number of wrongly declared schema items in the core
#
# Revision 1.1  2008/07/17 16:40:55  moscicki
# migration of 5.0.2 to HEAD
#
# the doc and release/tools have been taken from HEAD
#
# Revision 1.15.26.8  2008/05/15 06:31:38  moscicki
# # bugfix 36398: allow to assign a list in the configuration
#
# Revision 1.15.26.7  2008/04/18 13:42:18  moscicki
# remove obsolete printout
#
# Revision 1.15.26.6  2008/04/18 10:52:13  moscicki
# 1) typechecking fix
# 2) Ganga/test/GPI/ConfigSetEmptyVOMSString.gpi fix
#
# Revision 1.15.26.5  2008/04/18 08:14:38  moscicki
# bugfix 18272 (reintroduced in Ganga 5): add typelist information to the configuration option
#
# Revision 1.15.26.4  2007/12/18 09:07:06  moscicki
# integrated typesystem from Alvin
#
# Revision 1.15.26.3  2007/11/07 17:02:13  moscicki
# merged against Ganga-4-4-0-dev-branch-kuba-slices with a lot of manual merging
#
# Revision 1.15.26.2  2007/11/07 15:10:03  moscicki
# merged in pretty print and GangaList support from ganga-5-dev-branch-4-4-1-will-print branch
#
#
# Revision 1.15.26.1  2007/10/12 13:56:24  moscicki
# merged with the new configuration subsystem
#
# Revision 1.15.28.1  2007/09/25 09:45:11  moscicki
# merged from old config branch
#
# Revision 1.15.6.1  2007/06/18 07:44:55  moscicki
# config prototype
#
# Revision 1.15.30.1  2007/10/30 12:12:08  wreece
# First version of the new print_summary functionality. Lots of changes, but some known limitations. Will address in next version.
#
# Revision 1.15.8.1  2007/06/18 10:16:36  moscicki
# slices prototype
#
# Revision 1.15  2007/03/05 12:04:18  moscicki
# explicit switch for strict_sequence (default is True), if the sequence is non-strict then a single value v will be converted to [v] on assignment, for example non-strict File sequence yields obj.x = 'a' <=> obj.x = [File('a')]  <=> obj.x = File('a')
#
# Revision 1.14  2006/10/02 13:11:18  moscicki
# added extra check when setting default values: try to apply the value and raise ConfigError if it fails (for example unknown plugin)
#
# Revision 1.13  2006/07/28 12:53:36  moscicki
# fixed default value for ComponentItem (was broken)
#
# Revision 1.12  2006/07/28 08:26:23  moscicki
# allow defvalue to be an object for component items (not only a string or None)
#
# Revision 1.11  2006/07/27 20:15:35  moscicki
# createDefaultConfig()
# getDefaultValue()
# "checkset" metaproperty
#
# Revision 1.10  2005/12/02 15:35:06  moscicki
# visitable and getter metaproperties
#
# Revision 1.9  2005/08/24 15:42:21  moscicki
# automatically generated help for properties, disabled the SchemaHelper and few other improvements to the help system
#
#
#



################################################################################
# GangaND280 Project.
# Anthony Hillairet
# Created 26/12/2013
################################################################################
"""@package Highland
This module is designed to run any highland executable accessible in the $PATH environment variable.
"""


from GangaCore.GPIDev.Adapters.IPrepareApp import IPrepareApp
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Schema import *

from GangaCore.Utility.Config import getConfig

from GangaCore.GPIDev.Lib.File import *
from GangaCore.GPIDev.Lib.Registry.PrepRegistry import ShareRef
from GangaCore.GPIDev.Base.Proxy import isType
from GangaCore.Core.exceptions import ApplicationConfigurationError

import os, shutil, re
from GangaCore.Utility.files import expandfilename
shared_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),'shared',getConfig('Configuration')['user'])

class Highland(IPrepareApp):
    """
    Highland application running any highland executables.

    The required input for this module are:
        app.exe = 'RunNumuCCAnalysis.exe'
        app.cmtsetup = '/home/me/myT2KWork/setup.sh'
        app.outputfile = 'myFantasticResults.root'

    NOTE: Ganga will run a bash script so the CMT setup script must be in sh/bash.

    It is also possible to pass command line arguments to the executable.
    For example to analyze only 10 events:
        app.args = ['-n',10]
    
    """
    _schema = Schema(Version(3,3), {
        'exe' : SimpleItem(defvalue=None,typelist=['str', 'type(None)'],comparable=1,doc='A path (string) or a File object specifying an executable.'), 
        'args' : SimpleItem(defvalue=[],typelist=['str','GangaCore.GPIDev.Lib.File.File.File','int'],sequence=1,strict_sequence=0,doc="List of arguments for the executable. Arguments may be strings, numerics or File objects."),
        'cmtsetup' : SimpleItem(defvalue=None,doc='Setup script in bash to set up cmt and the cmt package of the executable.', typelist=['str','type(None)']),
        'outputfile' : SimpleItem(defvalue=None,doc='Output file name.', typelist=['str','type(None)']),
        'env' : SimpleItem(defvalue={},typelist=['str'],doc='Environment'),
        'is_prepared' : SimpleItem(defvalue=None, strict_sequence=0, visitable=1, copyable=1, typelist=['type(None)','bool'],protected=0,comparable=1,doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
        } )
    _category = 'applications'
    _name = 'Highland'
    _scriptname = None
    _exportmethods = ['prepare']
    _GUIPrefs = [ { 'attribute' : 'exe', 'widget' : 'File' },
                  { 'attribute' : 'args', 'widget' : 'String_List' },
                  { 'attribute' : 'outputfile', 'widget' : 'String' },
                  { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    _GUIAdvancedPrefs = [ { 'attribute' : 'exe', 'widget' : 'File' },
                          { 'attribute' : 'args', 'widget' : 'String_List' },
                          { 'attribute' : 'outputfile', 'widget' : 'String' },
                          { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    def __init__(self):
        super(Highland,self).__init__()


    def configure(self,masterappconfig):
        
        self.args = convertIntToStringArgs(self.args)

        job = self.getJobObject()

        if self.cmtsetup is None:
          raise ApplicationConfigurationError('No cmt setup script given.')

        # setup the output file
        for arg in self.args:
          if arg == '-o':
            raise ApplicationConfigurationError('Option "-o" given in args. You must use the outputfile variable instead.')

        if self.outputfile is None:
          raise ApplicationConfigurationError('No output file given. Fill the outputfile variable.')
        else:
          self.args.append('-o')
          self.args.append(self.outputfile)


        # So get the list of filenames get_dataset_filenames() and create a file containing the list of files and put it in the sandbox
        fileList = job.inputdir+'FileList'
        if not job.inputdata.set_dataset_into_list(fileList):
          raise ApplicationConfigurationError('Problem with the preparation of the list of input files')
        self.args.append(fileList)

        argsStr = ' '.join(self.args)
        # ANT: Create the bash script here and put it in input dir.
        script = '#!/bin/bash\n'
        script += 'source '+self.cmtsetup+'\n'
        script += self.exe+' '+argsStr+'\n'

        from GangaCore.GPIDev.Lib.File import FileBuffer

        if self.exe.find('.exe') > -1:
            scriptname = self.exe.replace('.exe', '.sh')
        else:
            scriptname = self.exe + '.sh'
        job.getInputWorkspace().writefile(FileBuffer(scriptname,script),executable=1)

        self._scriptname = job.inputdir+scriptname

        return (None,None)



config = getConfig('defaults_Highland') #_Properties
config.options['exe'].type = type(None)


def convertIntToStringArgs(args):

    result = []
    
    for arg in args:
        if isinstance(arg,int):
            result.append(str(arg))
        else:
            result.append(arg)

    return result



class RTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

        return StandardJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)
        

class LCGRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.LCG import LCGJobConfig

        return LCGJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

class gLiteRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.gLite import gLiteJobConfig

        return gLiteJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

allHandlers.add('Highland','LSF', RTHandler)
allHandlers.add('Highland','Local', RTHandler)
allHandlers.add('Highland','PBS', RTHandler)
allHandlers.add('Highland','SGE', RTHandler)
allHandlers.add('Highland','Slurm', RTHandler)
allHandlers.add('Highland','Condor', RTHandler)
allHandlers.add('Highland','LCG', LCGRTHandler)
allHandlers.add('Highland','gLite', gLiteRTHandler)
allHandlers.add('Highland','TestSubmitter', RTHandler)
allHandlers.add('Highland','Interactive', RTHandler)
allHandlers.add('Highland','Batch', RTHandler)
allHandlers.add('Highland','Cronus', RTHandler)
allHandlers.add('Highland','Remote', LCGRTHandler)
allHandlers.add('Highland','CREAM', LCGRTHandler)





#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
'''Common XMLPostProcessing code for AppBase, Bender and GaudiPython objects for
   LHCb. This is common to these jobs but relies on the LHCb output information.'''

# Required for post-processing script
import os
import sys
import re
from GangaLHCb.Lib.Applications.AppsBaseUtils import backend_handlers, activeSummaryItems

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def _XMLJobFiles():
    return ['summary.xml', '__parsedxmlsummary__']

# Post-Processing script taken from AppBase to be shared among multiple
# Job types

def postprocess(self, logger):

    j = self.getJobObject()
    parsedXML = os.path.join(j.outputdir, '__parsedxmlsummary__')
    # use to avoid replacing 'lumi' etc as return value and not the method
    # pointer
    metadataItems = {}
    if os.path.exists(parsedXML):
        #Get rid of the long representation
        xml_string = open(parsedXML).read()
        xml_string = re.sub('(\d)L([\,\}])', r'\1\2', xml_string)
        exec(compile(xml_string, parsedXML, 'exec'), {}, metadataItems)

    # Combining subjobs XMLSummaries.
    if j.subjobs:
        env = self.getenv(self.is_prepared is None)
        if 'XMLSUMMARYBASEROOT' not in env:
            logger.warning(
                '"XMLSUMMARYBASEROOT" env var not defined so summary.xml files not merged for subjobs of job %s' % j.fqid)
            return

        summaries = []
        for sj in j.subjobs:
            outputxml = os.path.join(sj.outputdir, 'summary.xml')
            if not os.path.exists(outputxml):
                logger.warning("XMLSummary for job %s subjobs will not be merged as 'summary.xml' not present in job %s outputdir" % (j.fqid, sj.fqid))
                return
            elif os.path.getsize(outputxml) == 0 or os.stat(outputxml).st_size == 0:
                logger.warning("XMLSummary for job %s subjobs will not be merged as %s appears to be an empty file" % (j.fqid, outputxml))
                logger.warning("Please try to recreate this file by either resubmitting your job or re-downloading the data from the backend")
                return
            summaries.append(outputxml)

        # Not needed now that we dont merge if ANY of subjobs have missing summary.xml
        #if not summaries:
        #    logger.debug('None of the subjobs of job %s produced the output XML summary file "summary.xml". Merging will therefore not happen' % j.fqid)
        #    return

        schemapath = os.path.join(env['XMLSUMMARYBASEROOT'], 'xml/XMLSummary.xsd')
        from GangaLHCb.Lib.XMLSummary.summary import Merge

        try:
            XMLSummarydata = Merge(summaries, schemapath)
        except Exception as err:
            logger.error('Problem while merging the subjobs XML summaries')
            raise

        for name, method in activeSummaryItems().items():
            try:
                metadataItems[name] = method(XMLSummarydata)
            except:
                metadataItems[name] = None
                logger.debug('Problem running "%s" method on merged xml output.' % name)

    for key, value in metadataItems.items():
        if value is None:  # Has to be explicit else empty list counts
            j.metadata[key] = 'Not Available.'
        else:
            j.metadata[key] = value

def GaudiExecPostProcess(self, logger):
    from GangaGaudi.Lib.Applications import GaudiXMLSummary
    j = self.getJobObject()
    metadataItems = {}
    #If this is a subjob then get the metadata from the summary.xml file
    if os.path.exists(os.path.join(j.outputdir, 'summary.xml')):
       sjSummary =  GaudiXMLSummary(j, 'summary.xml').summary()
       sjMetadataItems = {}
       for name, method in activeSummaryItems().items():
           try:
               sjMetadataItems[name] = method(sjSummary)
           except:
               sjMetadataItems[name] = None
               logger.debug('Problem running "%s" method on merged xml output.' % name)

       for key, value in sjMetadataItems.items():
           if value is None:  # Has to be explicit else empty list counts
               j.metadata[key] = 'Not Available.'
           else:
               j.metadata[key] = value

    # Combining subjobs XMLSummaries.
    if j.subjobs:
        env = self.getenv(self.is_prepared is None)
        if 'XMLSUMMARYBASEROOT' not in env:
            logger.warning(
                '"XMLSUMMARYBASEROOT" env var not defined so summary.xml files not merged for subjobs of job %s' % j.fqid)
            return

        summaries = []
        for sj in j.subjobs:
            outputxml = os.path.join(sj.outputdir, 'summary.xml')
            if not os.path.exists(outputxml):
                logger.warning("XMLSummary for job %s subjobs will not be merged as 'summary.xml' not present in job %s outputdir" % (j.fqid, sj.fqid))
                return
            elif os.path.getsize(outputxml) == 0 or os.stat(outputxml).st_size == 0:
                logger.warning("XMLSummary fro job %s subjobs will not be merged as %s appears to be an empty file" % (j.fqid, outputxml))
                logger.warning("Please try to recreate this file by either resubmitting your job or re-downloading the data from the backend")
                return
            summaries.append(outputxml)

        # Not needed now that we dont merge if ANY of subjobs have missing summary.xml
        #if not summaries:
        #    logger.debug('None of the subjobs of job %s produced the output XML summary file "summary.xml". Merging will therefore not happen' % j.fqid)
        #    return

        schemapath = os.path.join(env['XMLSUMMARYBASEROOT'], 'xml/XMLSummary.xsd')
        from GangaLHCb.Lib.XMLSummary.summary import Merge

        try:
            XMLSummarydata = Merge(summaries, schemapath)
        except Exception as err:
            logger.error('Problem while merging the subjobs XML summaries')
            raise

        for name, method in activeSummaryItems().items():
            try:
                metadataItems[name] = method(XMLSummarydata)
            except:
                metadataItems[name] = None
                logger.debug('Problem running "%s" method on merged xml output.' % name)

    for key, value in metadataItems.items():
        if value is None:  # Has to be explicit else empty list counts
            j.metadata[key] = 'Not Available.'
        else:
            j.metadata[key] = value




from GangaCore.GPIDev.Adapters.ISplitter import ISplitter
from GangaCore.GPIDev.Schema import *
from GangaCore.GPIDev.Lib.Job import Job


class TestSplitter(ISplitter):

    '''Split a job into n subjobs
    '''
    _name = "TestSplitter"
    docstr = "number of subjobs"
    _schema = Schema(Version(1, 0),
                     {'numJobs': SimpleItem(defvalue=10, doc=docstr)})

    def split(self, job):
        subjobs = []

        for i in range(self.numJobs):
            j = Job()
            j.copyFrom(job)
            j.splitter = None
            j.merger = None
            j.inputsandbox = []  # master added automatically
            subjobs.append(j)

        return subjobs



#!/usr/bin/env python

import sys
import re
import tempfile
import os
import optparse
import shutil
import random

from GangaCore.Utility.logging import getLogger
logger = getLogger()

class ND280Config:


    """
    This is an adaptation of ND280Computing/tools/ND280Configs.py.
    A class to handle automatically creating nd280Control config info.

    Types

    gnsetup = GENIE setup config
    raw     = Raw data config

    Defaults

    General

    cmtpath = environment
    cmtroot = environment

    raw

    module_list              = oaCalib oaRecon oaAnalysis

    An example use:

    >>> import ND280Configs
    >>> cfg = ND280Configs.ND280Config('raw')
    >>> cfg.ListOptions()
comment = v11r31-wg-bugaboo
cmtpath = environment
midas_file = /neut/datasrv2a/vavilov/nd280_00005012_0033.daq.mid.gz
num_events =
save_geometry = 1
enable_modules =
cmtroot = environment
db_time = 2014-06-25 06:00:00
module_list = oaCalib oaRecon oaAnalysis
version_number =
event_select = SPILL
custom_list =
nd280ver = v11r31
production = True
process_truncated =
disable_modules =
inputfile =

    >>> cfg.options['nd280ver']='v11r31'
    >>> cfg.options['comment'] = 'v11r31-wg-bugaboo'
    >>> cfg.options['db_time']='2014-06-25 06:00:00'

    >>> cfg.options['event_select'] = 'SPILL'
    >>> cfg.options['midas_file'] = '/global/scratch/t2k/raw/ND280/ND280/00010000_00010999/nd280_00010290_0000.daq.mid.gz'

    >>> cfg.options['production'] = 'True'
    >>> cfg.options['save_geometry'] = '1'

    >>> cfg_str = cfg.CreateConfig()
    >>> print cfg_str # prints configuration file info
    """

    def __init__(self, cfgtype,extra_opts):
        self.cfgtype=cfgtype.lower()
        self.options=self.common_options
        self.options_ignore=self.common_options_ignore

        if self.cfgtype == 'gnsetup':
            self.options.update(self.genieSetupt_options)
            self.options['genie_setup_script']=self.t2ksoftdir + '/GENIE/setup.sh'
        elif self.cfgtype == 'gnmc':
            self.options.update(self.genieMC_options)
            self.options['genie_setup_script']=self.t2ksoftdir + '/GENIE/setup.sh'
        elif self.cfgtype == 'raw':
            self.options.update(self.raw_options)
            self.options_ignore.update(self.raw_options_ignore)
        elif self.cfgtype == 'cosmicmc':
            self.options.update(self.cosmicmc_options)
            self.options_ignore.update(self.cosmicmc_options_ignore)
        elif self.cfgtype == 'sandmc':
            self.options.update(self.sandmc_options)
            self.options_ignore.update(self.sandmc_options_ignore)
        elif self.cfgtype == 'gncp':
            self.options.update(self.gncp_options)
        elif self.cfgtype == 'MC':
            self.options.update(self.mc_options)
        elif self.cfgtype == 'PG':
            self.options.update(self.pg_options)
        else:
            raise 'ND280Configs: Not a recognised config type'

        self.options.update(extra_opts)

    ### Global
    cfgtype=''
    config_filename=''
    t2ksoftdir=''
    options={}
    ##################
    ### Options
    ### Common Options: Dictionary of common options amongst all config files
    common_options={'cmtpath':'environment',
                    'cmtroot':'environment',
                    'nd280ver':'',
                    'custom_list':'',
                    'db_time':'',
                    'database_P6':''}

    ## Ignore options - options that are not essential and can be overlooked
    common_options_ignore={'comment':'','midas_file':'',
                           'event_select':'',
                           'inputfile':'',
                           'version_number':'',
                           'custom_list':'',
                           'db_time':'',
                           'database_P6':''}

    ### Raw Data Options: Dictionary of options specific to Raw data processing cfg files
    raw_options={ 'midas_file':'',
                  'comment':'',
                  'event_select':'',
                  'module_list':'oaCalib oaRecon oaAnalysis',
                  'version_number':'',
                  'inputfile':'',
                  'enable_modules':'',
                  'disable_modules':'',
                  'process_truncated':'',
                  'num_events':'',
                  'production':'',
                  'save_geometry':''}

    raw_options_ignore={'enable_modules':'',
                        'disable_modules':'',
                        'process_truncated':'',
                        'num_events':'',
                        'production':'',
                        'save_geometry':''}

    cosmicmc_options={   'stage':'base',
                         'nd280ver':'v11r31',
                         'module_list':'cosmic nd280MC elecSim oaCalibMC oaCosmicTrigger',
                         'run_number':'0',
                         'subrun':'0',
                         'baseline':'2010-11',
                         'p0d_water_fill':'1',
                         'replace_comment':'1',
                         'comment':'basecosmiccorsika5F',
                         'mc_type':'Cosmic',
                         'random_seed':'123456789',
                         'num_events':'1000000000',
                         'mc_full_spill':'0',
                         'mc_position':'Free',
                         'kinfile':'REPLACE_KINFILE',
                         'inputfile':'REPLACE_INPUTFILE',
                         'randomize_kinfile':'false',
                         'save_digits':'true' }

    cosmicmc_options_ignore={'random_seed':''}

    #### Sand MC
    sandmc_options={'stage':'neutMC',
                'module_list':'sandPropagate nd280MC elecSim oaCalibMC oaRecon oaAnalysis',
                'neut_setup_script':'REPLACE_NEUTSETUP',
                'neut_card':'REPLACE_NEUTCARD',
                'flux_file':'REPLACE_FLUXFILE',
                'maxint_file':'REPLACE_MAXINT',
                'inputfile':'REPLACE_INPUTFILE',
                'run_number':'0',
                'subrun':'0',
                'baseline':'2010-11',
                'p0d_water_fill':'0',
                'num_events':'10000000',
                'mc_type':'Neut_RooTracker',
                'nd280mc_random_seed':'3456532423',
                #'nd280mc_random_seed':str(random.getrandbits(29)),
                'pot':'2.5e17',
                'neutrino_type':'beam',
                'flux_region':'SAND',
                'master_volume':'World',
                'force_volume_name':'true',
                'tpc_periods_to_activate':'runs2-3',
                'ecal_periods_to_activate':'3-4',
                'random_start':'1',
                'nbunches':'8',
                'interactions_per_spill':'1',
                'pot_per_spill':'1',
                'mc_full_spill':'1',
                'time_offset':'50',
                'count_type':'MEAN',
		'mc_position':'free',
                'bunch_duration':'19',
                'elmc_random_seed':'317075105',
                #'elmc_random_seed':str(random.getrandbits(28)),
                'production':'True',
                'save_geometry':'1',
                'comment':'prod6sand201011airrun3'}

    sandmc_options_ignore={}

    #### MC
    mc_options={'module_list':'nd280MC elecSim oaCalibMC oaRecon oaAnalysis',
                'inputfile':'/lustre/ific.uv.es/sw/t2k.org/nd280Soft/nd280computing/processing_scripts/oa_nt_beam_90210013-0100_3ravbul66rum_numc_000_prod004magnet201011waterb.root',
                'run_number':'90210000',
                'subrun':'0',
                'baseline':'2010-11',
                'p0d_water_fill':'1',
                'num_events':'100000000',
                'mc_type':'Neut_RooTracker',
#                'nd280mc_random_seed':'3456532423',
                'nd280mc_random_seed':str(random.getrandbits(29)),
                'nbunches':'8',
                # for production 5
                'interactions_per_spill':'9.2264889',
                'pot_per_spill':'7.9891e+13',
                'mc_full_spill':'1',
                'time_offset':'50',
                'count_type':'MEAN',
		'mc_position':'free',
                'bunch_duration':'19',
#                'elmc_random_seed':'317075105',
                'elmc_random_seed':str(random.getrandbits(28)),
                'production':'1',
                'comment':'prod004magnet201011waterb'}

    ### GENIE Setup Options: Dictionary of options specific to GENIE setup cfg files
    genieSetup_options={'baseline':'',
                        'p0d_water_fill':'',
                        'genie_xs_table':'',
                        'master_volume':'',
                        'genie_setup_script':''}

    ### GENIE MC Options: Dictionary of options specific to GENIE MC cfg files
    genieMC_options={'baseline':'',
                     'p0d_water_fill':'',
                     'genie_xs_table':'',
                     'master_volume':'',
                     'genie_setup_script':'',
                     'run_number':'',
                     'subrun':'',
                     'comment':'',
                     'neutrino_type':'beam',
                     'flux_file':'',
                     'flux_tree':'h3002',
                     'pot':'',
                     'genie_paths':'',
                     'random_seed':''}

    gncp_options = {'module_list':'oaCherryPicker nd280MC elecSim oaCalibMC oaRecon oaAnalysis',
                    'run_number':'',
                    'subrun':'',
                    'comment':'',
                    'baseline':'',
                    'p0d_water_fill':'',
                    'num_events':'999999999',
                    'nd280mc_random_seed':'',
                    'mc_full_spill':'0',
                    'time_offset':'50',
                    'mc_position':'free',
                    'interactions_per_spill':'1',
                    'count_type':'FIXED',
                    'elmc_random_seed':'',
                    'cherry_picker_type':''}

    ### GRID options are not considered compulsory but I include here for totality.
    grid_options = {'use_grid':'0',
                    'storage_address':'',
                    'register_address':'',
                    'register_files':'1',
                    'register_catalogue_files':'0'}

    ### Particle Gun options
    pg_options = {'geo_baseline':'2010-02',
                  'p0d_water':'1',
                  'module_list':'nd280MC elecSim oaCalibMC oaRecon oaAnalysis',
                  'comment':'',
                  'num_events':'10000',
                  'mc_particle':'mu-',
                  'mc_position':'SUBDETECTOR p0d',
                  'mc_energy':'uniform 500 1000',
                  'mc_direction':'ISOTROPIC',
                  'nd280mc_random_seed':'999',
                  'interactions_per_spill':'1',
                  'nbunches':'1',
                  'elmc_random_seed':'999'}


    def CheckOptions(self):
        allOK=1
        ## Quick check to see if there are any blank options, only allowed for comment.
        for k,v in self.options.items():
            if not v:
                if k in self.options_ignore:
                    continue
                else:
                    logger.info('Please ensure a value for ' + k + ' using the object.options[\'' + k + '\']')
                    allOK=0

        return allOK

    def ListOptions(self):
        for k,v in self.options.items():
            logger.info(k + ' = ' + v)
        return

    def SetOptions(self,options_in):
        for k,v in options_in.items():
            if k in self.options:
                self.options[k]=v
            else:
                logger.info('Option ' + k + ' not in list, ignoring.')

    def CreateConfig(self):
        map = {
            'gnsetup' : self.CreateGENIEsetupCF,
            'raw'     : self.CreateRawCF,
            'cosmicmc': self.CreateCosmicMCCF,
            'sandmc': self.CreateSandMCCF
            }
        if map.get(self.cfgtype):
            creator = map[self.cfgtype]
            return creator()
        else:
            raise 'ND280Configs: Not a reconised config type'



    ######################## GENIE Setup
    def CreateGENIEsetupCF(self):
        pass


    ######################## Raw Data Processing Config file
    def CreateRawCF(self):

        if not self.CheckOptions():
            logger.error('Please make sure all options stated above are entered')
            return ''

        configfile = ''
        configfile += "# Automatically generated config file\n\n"

        ### Software Setup
        configfile += "[software]\n"
        configfile += "cmtpath = " + self.options['cmtpath'] + "\n"
        configfile += "cmtroot = " + self.options['cmtroot'] + "\n"
        configfile += "nd280ver = " + self.options['nd280ver'] + "\n\n"

        ### File naming
        if not  self.options['comment']:
             self.options['comment'] =  self.options['nd280ver']
        configfile += "[filenaming]\n"
        if self.options['inputfile']:
            logging.info("version_number = " + self.options['version_number'] + "\n")
            configfile += "version_number = " + self.options['version_number'] + "\n"
        configfile += "comment = " + self.options['comment'] + "\n\n"


        ### Module list
        configfile += "[configuration]\n"
        if not self.options['midas_file']:
            configfile += "inputfile = " + self.options['inputfile'] + "\n"
        configfile += "module_list = " + self.options['module_list'] + "\n"

        if self.options['db_time']:
            configfile += "database_rollback_date = " + self.options['db_time'] + "\n"
            configfile += "dq_database_rollback_date = " + self.options['db_time'] + "\n"

        if self.options['database_P6']:
            configfile += 'database_P6 = %s\n' % self.options['database_P6']

        configfile += "\n"

        ### Unpack - only if this is a midas file
        if self.options['midas_file']:
            configfile += "[unpack]\n"
            configfile += "midas_file = " + self.options['midas_file'] + "\n"
            configfile += "event_select = " + self.options['event_select'] + "\n"

            if self.options['process_truncated']:
                configfile += "process_truncated = " + self.options['process_truncated'] + "\n"
                configfile += "\n"
            if self.options['num_events']:
                configfile += "num_events = " + self.options['num_events'] + "\n"
            configfile += "\n"

        ### Analysis if corresponding options are present
        if self.options['enable_modules'] or self.options['disable_modules'] or \
               self.options['production'] or self.options['save_geometry']:
            configfile += "[analysis]\n"
            if self.options['enable_modules']:
                configfile += "enable_modules = " + self.options['enable_modules'] + "\n"
            if self.options['disable_modules']:
                configfile += "disable_modules = " + self.options['disable_modules'] + "\n"
            if self.options['production']:
                 configfile += "production = " +  self.options['production'] + "\n"
            if self.options['save_geometry']:
                 configfile += "save_geometry = " +  self.options['save_geometry'] + "\n"
            configfile += "\n"


        ### GRID Tools
        if 'use_grid' in self.options and self.options['use_grid']=='1':
            configfile += "[grid_tools]\n"
            configfile += "use_grid = " + self.options['use_grid'] + "\n"
            configfile += "storage_address = " + self.options['storage_address'] + "\n"
            configfile += "register_files = " + self.options['register_files'] + "\n"
            configfile += "register_catalogue_files = " + self.options['register_catalogue_files'] + "\n"
            configfile += "register_address = " + self.options['register_address'] + "\n"

        ### Custom options
        if self.options['custom_list']:
            configfile += '\n'
            configlist = self.options['custom_list'].split(',')
            for confline in configlist:
                configfile += str(confline)+"\n"
            configfile += '\n'

        return configfile

    ######################## Cosmic Processing Config file
    def CreateCosmicMCCF(self):

        if not self.CheckOptions():
            logging.error('Please make sure all options stated above are entered')
            return ''

        if not self.options['stage'] in ['base','fgd','tript','all']:
            logging.error('"stage" options should be one of',['base','fgd','tript','all'])
            return ''

        configfile = ''
        configfile += "# Automatically generated config file\n\n"


        ### Software Setup
        configfile += "[software]\n"
        configfile += "cmtpath = " + self.options['cmtpath'] + "\n"
        configfile += "cmtroot = " + self.options['cmtroot'] + "\n"
        configfile += "nd280ver = " + self.options['nd280ver'] + "\n\n"

        ### Module list
        configfile += "[configuration]\n"
        if self.options['stage'] != 'base':
            configfile += "module_list = oaRecon oaAnalysis\n"
            configfile += "inputfile = " +  self.options['inputfile'] + "\n\n"
        else:
            configfile += "module_list = " + self.options['module_list'] + "\n\n"

        ### File naming
        if not  self.options['comment']:
             self.options['comment'] =  self.options['nd280ver']
        configfile += "[filenaming]\n"
        configfile += "run_number = " + self.options['run_number'] + "\n"
        configfile += "subrun = " + self.options['subrun'] + "\n"
        configfile += "replace_comment = " + self.options['replace_comment'] + "\n"
        configfile += "comment = " + self.options['stage'] + "cosmiccorsika5F\n\n"


        ### Geometry
        configfile += "[geometry]\n"
        configfile += "baseline = " + self.options['baseline'] + "\n"
        configfile += "p0d_water_fill = " + self.options['p0d_water_fill'] + "\n\n"

        ### nd280mc
        configfile += "[nd280mc]\n"
        configfile += "mc_type = " + self.options['mc_type'] + "\n"
        configfile += "random_seed = " + str(random.randint(1,999999999)) + "\n"
        configfile += "num_events = " + self.options['num_events'] + "\n"
        configfile += "mc_full_spill = " + self.options['mc_full_spill'] + "\n"
        configfile += "mc_position = " + self.options['mc_position'] + "\n\n"

        ### Cosmic
        configfile += "[cosmics]\n"
        configfile += "kinfile = " + self.options['kinfile'] + "\n"
        configfile += "randomize_kinfile = " + self.options['randomize_kinfile'] + "\n\n"

        ### Electronics
        configfile += "[electronics]\n"
        configfile += "random_seed = " + str(random.randint(1,999999999)) + "\n\n"

         ### Calibrate
        configfile += "[calibrate]\n"
        configfile += "save_digits = " + self.options['save_digits'] + "\n\n"

         ### Reconstruction
        configfile += "[reconstruction]\n"
        if  self.options['stage'] != 'all':
            configfile += "event_select = " + self.options['stage'] + "cosmic\n"
        else:
            configfile += "event_select = all\n"

        return configfile

    ######################## Sand MC Processing Config file
    def CreateSandMCCF(self):

        if not self.CheckOptions():
            logging.error('Please make sure all options stated above are entered')
            return ''

        if not self.options['stage'] in ['neutMC','g4anal','neutSetup']:
            logging.error('"stage" options should be one of',['neutMC','g4anal','neutSetup'])
            return ''

        configfile = ''
        configfile += "# Automatically generated config file\n\n"


        ### Software Setup
        configfile += "[software]\n"
        configfile += "neut_setup_script = " + self.options['neut_setup_script'] + "\n\n"

        ### Module list
        configfile += "[configuration]\n"
        if self.options['stage'] == 'g4anal':
            configfile += "module_list = sandPropagate nd280MC elecSim oaCalibMC  oaRecon oaAnalysis\n"
            configfile += "inputfile = " +  self.options['inputfile'] + "\n\n"
        else:
            #configfile += "module_list = " + self.options['module_list'] + "\n\n"
            configfile += "module_list = neutMC\n\n"

        ### File naming
        if not  self.options['comment']:
             self.options['comment'] =  self.options['nd280ver']
        configfile += "[filenaming]\n"

        thisrun = 90007000
        if self.options['generator'] == "old-neut":
            thisrun += 2000000
        if self.options['generator'] == "anti-neut":
            thisrun = 80007000
        if self.options['generator'] == "genie" :
            thisrun += 1000000

        if self.options['beam'] == "beamc":
            thisrun += 300000
        else:
            logging.error("self.beam = " + self.beam + " is not supported!!!")
            return ''

        if self.options['p0d_water_fill']: # water
            thisrun += 10000

        configfile += "run_number = %d\n" % (thisrun+int(self.options['run_number']))
        configfile += "subrun = " + self.options['subrun'] + "\n"
        configfile += "comment = " + self.options['comment'] + "\n\n"


        ### Geometry
        configfile += "[geometry]\n"
        configfile += "baseline = " + self.options['baseline'] + "\n"
        configfile += "p0d_water_fill = " + self.options['p0d_water_fill'] + "\n\n"

        ### Neutrino
        configfile += "[neutrino]\n"
        configfile += "neut_card = " + self.options['neut_card'] + "\n"
        configfile += "flux_file = " + self.options['flux_file'] + "\n"
        configfile += "maxint_file = " + self.options['maxint_file'] + "\n"

        configfile += "pot = " + self.options['pot'] + "\n"
        #configfile += "num_events = 5000\n" # DVtmp
        configfile += "neutrino_type = " + self.options['neutrino_type'] + "\n"
        configfile += "flux_region = " + self.options['flux_region'] + "\n"
        configfile += "master_volume = " + self.options['master_volume'] + "\n"
        configfile += "force_volume_name = " + self.options['force_volume_name'] + "\n"
        configfile += "random_start = " + self.options['random_start'] + "\n"
        configfile += "random_seed = " + str(random.randint(1,999999999)) + "\n"
        configfile += "neut_seed1 = "  + str(random.randint(1,999999999)) + "\n"
        configfile += "neut_seed2 = "  + str(random.randint(1,999999999)) + "\n"
        configfile += "neut_seed3 = "  + str(random.randint(1,999999999)) + "\n\n"

        ### Dead channels
        configfile += "[dead_channels]\n"
        configfile += "tpc_periods_to_activate = " + self.options['tpc_periods_to_activate'] + "\n"
        configfile += "ecal_periods_to_activate = " + self.options['ecal_periods_to_activate'] + "\n\n"

        ### Sand propagate
        configfile += "[sandPropagate]\n"
        configfile += "num_events = " + self.options['num_events'] + "\n\n"

        ### nd280mc
        configfile += "[nd280mc]\n"
        configfile += "num_events = " + self.options['num_events'] + "\n"
        configfile += "mc_type = " + self.options['mc_type'] + "\n"
        configfile += "nbunches = " + self.options['nbunches'] + "\n"

        configfile += "interactions_per_spill = " + self.options['interactions_per_spill'] + "\n"
        configfile += "pot_per_spill = " + self.options['pot_per_spill'] + "\n"
        configfile += "bunch_duration = " + self.options['bunch_duration'] + "\n"
        configfile += "mc_full_spill = " + self.options['mc_full_spill'] + "\n"
        configfile += "time_offset = " + self.options['time_offset'] + "\n"
        configfile += "count_type = " + self.options['count_type'] + "\n"
        configfile += "mc_position = " + self.options['mc_position'] + "\n"
        configfile += "random_seed = " + str(random.randint(1,999999999)) + "\n\n"

        ### Electronics
        configfile += "[electronics]\n"
        configfile += "random_seed = " + str(random.randint(1,999999999)) + "\n\n"

        ### Analysis
        configfile += "[analysis]\n"
        configfile += "production = " +  self.options['production'] + "\n"
        configfile += "save_geometry = " +  self.options['save_geometry'] + "\n"

        return configfile

    ########################################################################################################################



################################################################################
# Ganga - a computational task management tool for easy access to Grid resources
# http://cern.ch/ganga
#
# $Id: runner.py,v 1.3 2008-11-26 08:31:33 moscicki Exp $
# runner.py is a Python module used to run Ganga test-cases
#
# Copyright (C) 2003-2007 The Ganga Project
#
# This file is part of GangaCore. 
#
# Ganga is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Ganga is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
################################################################################

__version__ = "1.0"
__author__="Adrian.Muraru[at]cern[dot]ch"

import os
import sys

## Configuration
from GangaCore.Utility.Config import getConfig, makeConfig
from GangaCore.Utility.files import previous_dir
from GangaCore.Utility.logging import getLogger

myFullPath =  os.path.abspath(os.path.dirname(__file__))
gangaReleaseTopDir = previous_dir(myFullPath,3)

# testing framework configuration properties
myConfig = makeConfig('TestingFramework','Configuration section for internal testing framework')

#Release testing mode:
# - xml/html reports are generated
# - coverage analysis tool enabled
myConfig.addOption('ReleaseTesting', False, '')

#enable/disable test-runner
myConfig.addOption('EnableTestRunner', True, 'enable/disable test-runner') 
#enable/disable html reporter
myConfig.addOption('EnableHTMLReporter', False, 'enable/disable html reporter')
#enable/disable xml differencer
myConfig.addOption('EnableXMLDifferencer','False', 'enable/disable xml differencer')
#search for local tests lookup
myConfig.addOption('SearchLocalTests', True, 'search for local tests lookup')
#search for tests packaged in release (PACKAGE/old_test dir) lookup
myConfig.addOption('SearchReleaseTests', True, 'search for tests packaged in release (PACKAGE/old_test dir) lookup')
#Coverage output
myConfig.addOption('CoverageReport', '', 'The file used to save the testing coverage statistics')

myConfig.addOption('timeout', 600, 'timeout when the test is forcibly stopped ')
myConfig.addOption('AutoCleanup',True,'cleanup the job registry before running the testcase')

myConfig.addOption('SchemaTesting', '', 'Set to True to enable Schema testing mode.')
myConfig.addOption('SchemaTest_ignore_obj', [], 'Objects to ignore when in Schema testing mode.')

if myConfig['SchemaTesting'] == '':
    myConfig.addOption('Config', 'default.ini', 'ganga configuration(s)')
else:
    myConfig.addOption('Config', 'Schema.ini', 'ganga configuration(s)')
    

if not myConfig['ReleaseTesting']:
    myConfig.addOption('OutputDir', os.path.expanduser('~/gangadir_testing'), '')
    myConfig.addOption('LogOutputDir', os.path.join(myConfig['OutputDir'],'output'), '')
    myConfig.addOption('ReportsOutputDir', os.path.join(myConfig['OutputDir'],'reports'), '')
    myConfig.addOption('RunID', '', '')
else:
    #for release testing
    myConfig.addOption('ReportsOutputDir', os.path.join(gangaReleaseTopDir,'reports'), '')
    myConfig.addOption('RunID', 'latest', '')
    myConfig.addOption('LogOutputDir', os.path.join(myConfig['ReportsOutputDir'],myConfig['RunID'],'output'), '')

#GangaTest logging config
loggerConfig = getConfig('Logging')
loggerConfig.addOption('GangaTest.Framework', "INFO", '')

#logger
myLogger=getLogger()

def start( config = myConfig , test_selection='GangaCore.test.*', logger=myLogger):
    """
    """    
    import os
    #rtconfig = getConfig('TestingFramework')
    my_full_path =  os.path.abspath(os.path.dirname(__file__))
    #sys.stdout = UnbufferedStdout(sys.stdout)
    #sys.stderr = UnbufferedStdout(sys.stderr)    
       
    ##configure Ganga TestLoader

    # enable XML reporting in release mode
    pytf_reporting_opts=""
    if config['ReleaseTesting']:
        pytf_reporting_opts="--report-xml --report-outputdir=%(ReportsOutputDir)s --runid=%(RunID)s" % config        

    # output dir 
    if not os.path.exists(config['LogOutputDir']):
        os.makedirs(config['LogOutputDir'])
        
    # loader path
    global gangaReleaseTopDir
    pytf_loader_path = os.path.join(gangaReleaseTopDir,'python','GangaTest','Framework','loader.py')
    
    # loader args
    pytf_loader_args =[]   
    pytf_loader_args.append( '--loader-args=%s' % config['Config'])
    #pytf_loader_args.append( '--loader-args=%s/python' %  gangaReleaseTopDir)
    pytf_loader_args.append( '--loader-args=%s' %  gangaReleaseTopDir)
    pytf_loader_args.append( '--loader-args=%s' % int(config['ReleaseTesting']))    
    #output_dir
    pytf_loader_args.append( '--loader-args=%s' % config['LogOutputDir'])
    #unit-testing: on/off
    pytf_loader_args.append( '--loader-args=%s' % int(config['SearchLocalTests'])) 
    #system-testing: on/off
    pytf_loader_args.append( '--loader-args=%s' % int(config['SearchReleaseTests'])) 
    #Pass the report(report path + runid) path
    pytf_loader_args.append( '--loader-args=%s' % os.path.join(config['ReportsOutputDir'],config['RunID']))
    #pass the schmema version (if any) to test
    pytf_loader_args.append( '--loader-args=%s' % config['SchemaTesting'])
    
    #print("PYTF path %s config: %s" % (pytf_loader_path, pytf_loader_args))
    import sys
    sys.path.append(os.getenv('PYTF_TOP_DIR','').split(':')[0])
    sys.path.append(os.path.join(os.getenv('PYTF_TOP_DIR','').split(':')[0],'pytf'))
    import runTests
    runner_args = []
    
    runner_args.extend(pytf_reporting_opts.split())
    runner_args.extend(['--loader-path=%s' % pytf_loader_path])
    runner_args.extend(pytf_loader_args)
    runner_args.extend([test_selection])

    try:
        rc = runTests.main(logger,runner_args)
    except:
        rc = -9999
    return rc
                

def __getExternalLibPath( current_path ):
    """
    Get the external lib path needed to start individual test-cases
    This can be achieved also by setting RUNTIME_PATH=GangaTest in the configuration
    but in case of local tests this cannot be enforced, so we set the PYTHONPATH manually
    """
    
    external_path = []
    ganga_external_dir = previous_dir(current_path,5)
   
    ganga_external_dir = os.path.join(ganga_external_dir,'external')    

    ##set the libraries path to include external packages and test directory
    pyth_path = os.path.join(ganga_external_dir,'PYTF',PYTF_VERSION,'slc3_gcc323')
    if not os.path.isdir(pyth_path):
        pyth_path = '/afs/cern.ch/sw/ganga/external/PYTF/%s/slc3_gcc323' % PYTF_VERSION      
    if pyth_path not in sys.path:
        sys.path.insert(0, pyth_path)
    external_path.append(pyth_path)
                    
    figleaf_path = os.path.join(ganga_external_dir,'figleaf',FIGLEAF_VERSION,'slc3_gcc323')
    if not os.path.isdir(figleaf_path):
        figleaf_path = '/afs/cern.ch/sw/ganga/external/figleaf/%s/slc3_gcc323' % FIGLEAF_VERSION
    if figleaf_path not in sys.path:
        sys.path.insert(0, figleaf_path)
    external_path.append(figleaf_path)
    return external_path
                

if __name__ == '__main__':
    start()
    
#$Log: not supported by cvs2svn $
#Revision 1.2  2008/09/05 09:53:18  uegede
#New features:
#- Remote testing of Ganga thorugh the use of the GangaRobot
#- Ability to create test reports showing differences between releases.
#
#Revision 1.6.4.6  2007/12/18 13:12:48  amuraru
#stream-line the test-case execution code (use a single entry-point : Framework/driver.py)
#symplified the interface with Ganga bootstrap
#
#Revision 1.6.4.5  2007/11/08 13:23:02  amuraru
#added CoverageReport option definition
#
#Revision 1.6.4.4  2007/11/02 14:47:24  moscicki
#moved options from utils.py
#
#Revision 1.6.4.3  2007/10/31 13:37:21  amuraru
#updated to the new config subsystem in Ganga 5.0
#
#Revision 1.6.4.2  2007/10/30 14:20:22  moscicki
#makeConfig imported
#
#Revision 1.6.4.1  2007/10/12 13:56:31  moscicki
#merged with the new configuration subsystem
#
#Revision 1.6.6.1  2007/10/09 07:23:23  roma
#Migration to new Config
#
#Revision 1.6  2007/09/04 09:48:13  amuraru
#restructed code to allow two modes of running: developer mode and release mode
#
#Revision 1.5  2007/08/13 12:33:15  amuraru
#added verbosity control, avoid purging the used reporisotry when running local tests
#
#Revision 1.4  2007/05/21 16:01:20  amuraru
#use default website css style in test/coverage reports;
#disabled per test-case coverage report generation;
#other fixes
#
#Revision 1.3  2007/05/16 10:15:52  amuraru
#use ganga logger
#
#Revision 1.2  2007/05/15 09:58:36  amuraru
#html reporter updated
#



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: __init__.py,v 1.1 2008-07-17 16:40:53 moscicki Exp $
##########################################################################
# File: Credentials/__init__.py
# Author: K.Harrison
# Created: 060613
#
# 08/08/2006 KH: Added getCredential() function
#
# 28/08/2006 KH: Don't determine available credentials from allPlugins,
#                as credential plugins are now defined as hidden
#
# 31/08/2006 KH: Corrections to getCredential() function to create
#                only a single instance of each credential type
#
# 23/11/2006 KH: Added check on credential availability with
#                system/configuration used
#
# 25/09/2007 KH:  Changes for compatibility with multi-proxy handling
#                 => "middleware" argument added to getCredential function
#
# 02/11/2007 KH:  Added argument to getCredential() function to allow
#                 or supress creation of new credential

"""Initialisation file for the Credentials package,
   containing classes for working with different types of credential."""

__author__ = "K.Harrison <Harrison@hep.phy.cam.ac.uk>"
__date__ = "25 September 2007"
__version__ = "1.5"

from GangaCore.Utility.logging import getLogger
from GangaCore.Utility.Plugin import allPlugins
from GangaCore.GPIDev.Credentials_old.GridProxy import GridProxy

_credentialPlugins = {}
for item in locals().keys():
    if ((hasattr(locals()[item], "_category"))
            and (hasattr(locals()[item], "_name"))):
        _category = getattr(locals()[item], "_category")
        if "credentials" == _category:
            _name = getattr(locals()[item], "_name")
            _credentialPlugins[_name] = locals()[item]
            del _name
        del _category

_allCredentials = {}
_voidCredentials = {}

logger = getLogger()


def getCredential(name="", create=True):
    """
    Function to return credential object of requested type

    Arguments:
       middleware - String specifying any middleware used with credential
       name       - String specifying credential type
       create     - Boole specifying:
                    True  - requested credential should be created if
                            it doesn't exist
                    False - no new credential should be created

    Return value: Credential object of requested type if it exists or
                  can be created, or False otherwise
    """

#  if name in allPlugins.allClasses( "credentials" ).keys():
#     if not name in _allCredentials.keys():
#        _allCredentials[ name ] = \
#           allPlugins.find( "credentials", name )._proxyClass()
    if name in _credentialPlugins.keys():
        if ( not name in list(_allCredentials.keys()) ) and \
           ( not name in list(_voidCredentials.keys()) ) and \
           (create is True):
            credential = _credentialPlugins[name]()
            if credential.isAvailable():
                _allCredentials[name] = credential
            else:
                _voidCredentials[name] = credential
    else:
        logger.warning("Credential type '%s' not known" % str(name))
        logger.warning("Returning False")

    if name in _allCredentials.keys():
        credential = _allCredentials[name]
    else:
        credential = False

    return credential



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: GridProxy.py,v 1.5 2009/03/25 15:43:35 karl Exp $
##########################################################################
#
# File: GridProxy.py
# Author: K. Harrison
# Created: 060519
#
# 06/07/2006 KH:  Changed to GangaCore.Utility.Shell for shell commands
#                 Added voms support
#
# 02/08/2006 KH:  Modified GridProxy class to create one instance of
#                 VomsCommand and GridCommand
#
# 07/08/2006 KH:  Added isValid() method
#
# 09/08/2006 KH:  Use shell defined via GangaCore.Lib.LCG.GridShell.getShell()
#
# 25/08/2006 KH:  Declare GridProxy class as hidden
#
# 06/09/2006 KH:  Argument minValidity added to methods create() and renew()
#
# 25/09/2006 KH:  Changed method isValid(), so that default validity is
#                 value of self.minValidity
#
# 13/11/2006 KH:  Added method info() for obtaining proxy information,
#                 and changed location() to use this method
#
# 23/11/2006 KH:  Added "pipe" keyword to option dictionaries of GridCommand
#                 and VomsCommand
#                 Added method to determine if credential is available
#                 with system/configuration used
#                 (requests from CLT)
#
# 28/02/2007 CLT: Replaced VomsCommand.options and GridCommand.options
#                 with dictionaries init_parameters, destroy_parameters,
#                 info_parameters, each providing independent options
#                 Added VomsCommand.currentOpts and GridCommand.currentOpts
#                 dictionaries, to add flexibility and assist in option
#                 construction (as opposed to direct string manipulation)
#                 Added GridProxy.buildOpts(), to consolidate the option
#                 building functionality from create(), destroy() and info()
#
# 02/03/2007 KH : Added method to determine user's identity
#                 (request from DL)
#
# 25/04/2007 KH : Modified GridProxy.identity method to be able to deal
#                 with new-style CERN certificates, with ambiguous CN definition
#
# 08/06/2007 KH : Added method GridProxy.voname, to allow name of
#                 virtual organisation to be determined from proxy
#
# 25/09/2007 KH:  Changes for compatibility with multi-proxy handling
#                 => "middleware" argument introduced
#
# 08/12/2007 KH:  Changes to take into account ICommandSet being made
#                 a component class
#
# 17/12/2007 KH:  Made changes for handling of GridCommand and VomsCommand as
#                 component classes
#
# 15/01/2008 KH:  Set initial null value for infoCommand in GridProxy.voname()
#
# 18/01/2008 KH : Modified GridProxy.identity method to disregard
#                 values of CN=proxy
#
# 27/02/2008 KH : Setup shell in GridProxy constructor, if middleware is defined
#
# 30/01/2009 KH : Added possibility to request that GridProxy.identity()
#                 returns string with non-alphanumeric characters stripped out
#
# 25/03/2009 KH : Correction to GridProxy.voname() to check that one-word
#                 VO name is returned
#
# 18/03/2009 MWS: Added the 'log' option to isValid
#                 Added method to retrieve the full identity as a dictionary
#                 Require consistency between VO in proxy and in configuration
#
# 15/10/2009 MWS: Added cache for proxy information
#
#
# 09/11/2009 MWS: Added check that proxy is valid before updating cache
#                 (addToProxyCache() method)
#
# 12/11/2009 MWS: Added additional checks that cached information
#                 is consistent with available proxy

"""Module defining class for creating, querying and renewing Grid proxy"""
__author__ = "K.Harrison <Harrison@hep.phy.cam.ac.uk>"
__date__ = "12 November 2009"
__version__ = "1.21"

import os
import re
import time

from GangaCore.GPIDev.Base.Proxy import isType
from GangaCore.GPIDev.Credentials_old.ICredential import ICommandSet, ICredential, registerCommandSet
from GangaCore.GPIDev.Schema import SimpleItem
from GangaCore.Utility.logging import getLogger
from GangaCore.Utility.GridShell import getShell

logger = getLogger()

allowed_exit_range = list(range(1000))


class GridCommand(ICommandSet):

    """
    Class used to define shell commands and options for working with Grid proxy
    """

    _schema = ICommandSet._schema.inherit_copy()
    _schema['init']._meta['defvalue'] = "grid-proxy-init"
    _schema['info']._meta['defvalue'] = "grid-proxy-info"
    _schema['destroy']._meta['defvalue'] = "grid-proxy-destroy"
    _schema['init_parameters']._meta['defvalue'] = {"pipe": "-pwstdin", "valid": "-valid"}
    _schema['destroy_parameters']._meta['defvalue'] = {}
    _schema['info_parameters']._meta['defvalue'] = {}

    _name = "GridCommand"
    _hidden = 1
    _enable_config = 1

    def __init__(self):
        super(GridCommand, self).__init__()

        self.currentOpts = {}
        self.infoOpts = {}
        self.destroyOpts = {}


class VomsCommand(ICommandSet):

    """
    Class used to define shell commands and options for working with Grid proxy,
    using VOMS extensions
    """

    _schema = ICommandSet._schema.inherit_copy()

    _schema['init']._meta['defvalue'] = "voms-proxy-init"
    _schema['info']._meta['defvalue'] = "voms-proxy-info"
    _schema['destroy']._meta['defvalue'] = "voms-proxy-destroy"
    _schema['init_parameters']._meta['defvalue'] = {"pipe": "-pwstdin", "valid": "-valid", "voms": "-voms"}
    _schema['destroy_parameters']._meta['defvalue'] = {}
    _schema['info_parameters']._meta['defvalue'] = {"vo": "-vo"}

    _name = "VomsCommand"
    _hidden = 1
    _enable_config = 1

    def __init__(self):
        super(VomsCommand, self).__init__()

        self.currentOpts = {}
        self.infoOpts = {}
        self.destroyOpts = {}

for commandSet in [GridCommand, VomsCommand]:
    registerCommandSet(commandSet)

# global proxy info cache
_infoCache = {}


class GridProxy(ICredential):

    """
    Class for working with Grid proxy
    """

    _schema = ICredential._schema.inherit_copy()
    _schema.datadict["voms"] = SimpleItem(defvalue="", doc="Virtual organisation managment system information")
    _schema.datadict["init_opts"] = SimpleItem(defvalue="", doc="String of options to be passed to command for proxy creation")
    _schema.datadict["info_refresh_time"] = SimpleItem(defvalue="00:15", doc="Refresh time of proxy info cache")
    _schema.datadict["maxTry"] = SimpleItem(defvalue=5, doc="Number of password attempts allowed when creating credential")
    _name = "GridProxy"
    _hidden = 1
    _enable_config = 1
    _exportmethods = ["create", "destroy", "identity", "info", "isAvailable",
                      "isValid", "location", "renew", "timeleft", "voname", "fullIdentity"]

    def __init__(self):
        super(GridProxy, self).__init__()
        self.shell = getShell()
        self.chooseCommandSet()
        return

    def chooseCommandSet(self):
        """
        Choose command set to be used for proxy-related commands

        No arguments other than self

        If self.voms has a null value then the GridCommand set of commands
        is used.  Otherwise the VomsCommand set of commands is used.

        Return value: None
        """
        from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
        if self.voms:
            # Make sure we don't overwrite a VomsCommand if it's already there
            if not isinstance(self.command, VomsCommand):
                self.command = VomsCommand()
        else:
            self.command = GridCommand()

    # Populate the self.command.currentOpts dictionary with
    # GridProxy specific options.
    def buildOpts(self, command, clear=True):
        if command == self.command.init:
            if clear:
                self.command.currentOpts.clear()
            if "voms" in self.command.init_parameters:
                if self.voms:
                    self.command.currentOpts\
                        [self.command.init_parameters['voms']] = self.voms
            if "valid" in self.command.init_parameters:
                if self.validityAtCreation:
                    self.command.currentOpts\
                        [ self.command.init_parameters[ 'valid' ] ] \
                        = self.validityAtCreation
            if self.init_opts:
                self.command.currentOpts[''] = self.init_opts
        elif command == self.command.destroy:
            if clear:
                self.command.destroyOpts.clear()
        elif command == self.command.info:
            if clear:
                self.command.infoOpts.clear()

    def create(self, validity="", maxTry=0, minValidity="", check=False):
        self.chooseCommandSet()
        self.buildOpts(self.command.init)
        status = ICredential.create(self, validity, maxTry, minValidity, check)
        return status

    def destroy(self, allowed_exit=[0, 1]):
        self.chooseCommandSet()
        self.buildOpts(self.command.destroy)
        return ICredential.destroy(self, allowed_exit)

    def isAvailable(self):
        if self.shell:
            return True
        else:
            return False

    def isValid(self, validity="", log=False, force_check=False):

        # Do parent check
        if not ICredential.isValid(self, validity, log, force_check):
            return False

        # check vo names
        if self.voname() != self.voms:
            if log:
                logger.warning("Grid Proxy not valid. Certificate VO '%s' does not match requested '%s'"
                               % (self.voname(), self.voms))
            return False

        return True

    def location(self):

        proxyPath = self.info("-path").strip()

        if not os.path.exists(proxyPath):
            proxyPath = ""

        return proxyPath

    def fullIdentity(self, safe=False):
        """
        Return the users full identity as a dictionary

        Argument:
           safe - logical flag
                  =>  False : return identity exactly as obtained from proxy
                  =>  True  : return identity after stripping out
                              non-alphanumeric characters

        Return value: Dictionary of the various labels in the users DN
        """

        ele_dict = {}

        subjectList = self.info(opt="-identity").split("/")

        for subjectElement in subjectList:
            element = subjectElement.strip()
            if element.find("=") == -1:
                continue

            field, val = element.split("=")
            if safe:
                val = re.sub("[^a-zA-Z0-9]", "", val)
            ele_dict[field] = val

        return ele_dict

    def identity(self, safe=False):
        """
        Return user's identify

        Argument:
           safe - logical flag
                  =>  False : return identity exactly as obtained from proxy
                  =>  True  : return identity after stripping out
                              non-alphanumeric characters

        => The identity is determined from the user proxy if possible,
           or otherwise from the user's top-level directory

        Return value: String specifying user identity
        """

        cn = os.path.basename(os.path.expanduser("~"))
        try:
            subjectList = self.info(opt="-identity").split("/")
            subjectList.reverse()
            for subjectElement in subjectList:
                element = subjectElement.strip()
                try:
                    cn = element.split("CN=")[1].strip()
                    if cn != "proxy":
                        break
                except IndexError:
                    pass
        except Exception as err:
            logger.debug("Err: %s" % err)
            pass

        id = "".join(cn.split())
        if safe:
            id = re.sub("[^a-zA-Z0-9]", "", id)

        return id

    def info(self, opt="", force_check=False):
        """
        Obtain proxy information

        Arguments other than self:
           opt         - String of options to be used when querying
                         proxy information
                       => Help on valid options can be obtained using:
                          info( opt = "-help" )
           force_check - Force credential check, rather than relying on cache

        Return value: Output from result of querying proxy
        """

        # use cached version of this command call if possible
        output = self.getProxyCacheValue(opt)

        if (force_check) or (output == ""):
            self.chooseCommandSet()
            infoCommand = " ".join([self.command.info, opt])
            logger.debug("Executing info Command: %s" % infoCommand)
            status, output, message = self.shell.cmd1(cmd=infoCommand, allowed_exit=allowed_exit_range)

            self.addToProxyCache(status, output, opt)

        if not output:
            output = ""

        return str("%s" % output)

    def renew(self, validity="", maxTry=0, minValidity="", check=True):
        self.chooseCommandSet()
        if self.voms:
            if not self.voname():
                check = False
        return ICredential.renew(self, validity, maxTry, minValidity, check)

    def timeleft(self, units="hh:mm:ss", force_check=False):
        return ICredential.timeleft(self, units, force_check)

    def timeleftInHMS(self, force_check=False):
        global _infoCache

        output = self.getProxyCacheValue("timeleftInHMS")
        status = 0

        if (force_check) or (output == ""):
            # should really use the 'info' method
            self.chooseCommandSet()
            infoList = [self.command.info]
            # Append option value pairs
            for optName, optVal in self.command.infoOpts.items():
                infoList.append("%s %s" % (optName, optVal))
            logger.debug("Executing timeHMS Command: %s" % " ".join(infoList))
            status, output, message = self.shell.cmd1\
                (cmd=" ".join(infoList), allowed_exit=allowed_exit_range)

            self.addToProxyCache(status, output, "timeleftInHMS")

        timeRemaining = "00:00:00"

        if status:
            if (1 + output.lower().find("command not found")):
                logger.warning("Command '" + self.command.info + "' not found")
                logger.warning("Unable to obtain information on Grid proxy")
                timeRemaining = ""
                if "timeleftInHMS" in _infoCache:
                    del _infoCache["timeleftInHMS"]

        if timeRemaining:
            lineList = output.split("\n")
            for line in lineList:
                if (1 + line.find("Couldn't find a valid proxy")):
                    timeRemaining = "-1"
                    if 'timeleftInHMS' in _infoCache:
                        del _infoCache['timeleftInHMS']
                    break
                elif (1 + line.find("timeleft")):
                    elementList = line.split()
                    timeRemaining = elementList[2]
                    break

        return timeRemaining

    def voname(self, force_check=False):
        """
        Obtain name of virtual organisation from proxy

        Argument other than self:
           force_check - Force credential check, rather than relying on cache

        Return value: Name of virtual organisation where this can be determined
        (voms proxy), or empty string otherwise (globus proxy)
        """
        global _infoCache
        output = self.getProxyCacheValue("voname")

        if (force_check) or (output == ""):
            self.chooseCommandSet()
            infoCommand = ""

            if "vo" in self.command.info_parameters:
                if self.command.info:
                    infoCommand = " ".join\
                        ([self.command.info,
                          self.command.info_parameters["vo"]])
            else:
                infoCommand = self.command.info

            if infoCommand:
                logger.debug("Executing voname Command: %s" % infoCommand)
                status, output, message = self.shell.cmd1(cmd=infoCommand,
                                                          allowed_exit=allowed_exit_range, capture_stderr=True)

                self.addToProxyCache(status, output, "voname")

            else:
                output = ""

        if not output:
            output = ""
            if "voname" in _infoCache:
                del _infoCache["voname"]

        output = output.strip()

        for error in ["VOMS extension not found", "unrecognized option"]:
            if output.find(error) != -1:
                output = ""
                if 'voname' in _infoCache:
                    del _infoCache['voname']
                break

        # Check for reasonable output (single-word VO)
        if len(output.split()) != 1:
            output = self.voms

        return output

    def getProxyCacheValue(self, opt):
        """
        Check the proxy cache for the required key. Make sure the proxy
        file is older than the last check.

        opt - the key to check for
        """

        global _infoCache

        info_refresh = self.timeInSeconds(self.info_refresh_time)
        output = ''
        path = ''

        # check when the grid proxy was created
        if '-path' not in _infoCache or (_infoCache['-path'][1] < (time.time() - info_refresh)):
            self.chooseCommandSet()
            infoCommand = " ".join([self.command.info, '-path'])
            logger.debug("Executing cache Command: %s" % infoCommand)
            status, output, message = self.shell.cmd1\
                (cmd=infoCommand, allowed_exit=allowed_exit_range)

            if not status:
                path = output
                self.addToProxyCache(status, output, '-path')

        else:
            path = _infoCache['-path'][0]

        path = path.strip()
        if not os.path.exists(path):
            # blank the cache as the proxy isn't there
            _infoCache = {}
            return ''

        # we're OK to use the cache
        if opt in _infoCache and\
            ( _infoCache[ opt ][ 1 ] > ( time.time() - info_refresh ) ) and\
                (_infoCache[opt][1] > os.path.getmtime(path)):
            logger.debug("Returning Cached Value %s" % opt)
            output = _infoCache[opt][0]
        else:
            output = ""

        return output

    def addToProxyCache(self, status, output, opt):
        """
        Test the result of grid proxy call
        and add to the cache if all OK

        status - the status output
        output - the output text
        opt - opt to add
        """

        if (not status) and (output):
            error = False
            for line in output.split('\n'):
                if (1 + line.find("Couldn't find a valid proxy")):
                    error = True

            if not error:
                _infoCache[opt] = [output, time.time()]

        return None

    # Add documentation strings from base class
    for method in [create, destroy, isAvailable, isValid, location,
                   renew, timeleft, timeleftInHMS]:
        if hasattr(ICredential, method.__name__):
            baseMethod = getattr(ICredential, method.__name__)
            setattr(method, "__doc__",
                    baseMethod.__doc__.replace("credential", "Grid Proxy"))



import os
import shutil
import time
import copy
import threading
from GangaCore.Core.GangaRepository.Registry import Registry
from GangaCore.GPIDev.Base import GangaObject
from GangaCore.GPIDev.Base.Objects import synchronised
from GangaCore.GPIDev.Schema import Schema, SimpleItem, Version
from GangaCore.GPIDev.Base.Proxy import stripProxy, getName
import GangaCore.Utility.Config
from GangaCore.GPIDev.Lib.File import getSharedPath
logger = GangaCore.Utility.logging.getLogger()


class PrepRegistry(Registry):

    def __init__(self, name, doc):

        super(PrepRegistry, self).__init__(name, doc)

    def startup(self):
        self._needs_metadata = True
        super(PrepRegistry, self).startup()
        this_ShareRef = ShareRef()
        if len(self.metadata.ids()) == 0:
            self.metadata._add(this_ShareRef)
        self.shareref = self.metadata[self.metadata.ids()[-1]]

    def getShareRef(self):
        return self.shareref

    def getSlice(self):
        pass

    def getProxy(self):
        pass

    def shutdown(self):
        """
        This allows us to clean up after the ShareDir on shutdown
        """
        if self.shareref:
            self.shareref.cleanUpOrphans()
        super(PrepRegistry, self).shutdown()

class ShareRef(GangaObject):

    """The shareref table (shared directory reference counter table) provides a mechanism
    for storing metadata associated with Shared Directories (see help(ShareDir)), which 
    may be referenced by other Ganga objects, such as prepared applications.
    When a Shared Directory is associated with a persisted Ganga object (e.g. Job, Box) its 
    reference counter is incremented by 1. Shared Directories with a reference counter of 0 will
    be removed (i.e. the directory deleted) the next time Ganga exits.
    """
    _schema = Schema(Version(1, 2), {'name': SimpleItem({}, protected=1, copyable=1, hidden=1)})

    _category = 'sharerefs'
    _name = 'ShareRef'
    _exportmethods = ['increase', 'decrease', 'counterVal', 'ls', 'printtree', 'rebuild', 'lookup', 'registerForRemoval']

    #_parent = None
    default_registry = 'prep'

    def __init__(self):
        super(ShareRef, self).__init__()
        self.name = {}
        self.removal_list = []

    def __setattr__(self, attr, value):
        actual_value = value
        if attr == "name":
            if not actual_value:
                actual_value = {}
        super(ShareRef, self).__setattr__(attr, actual_value)

    def __getName(self):
        if not hasattr(self, 'name') or self.name is None:
            self.name = {}
        return self.name

    @synchronised
    def registerForRemoval(self, shareddir):
        """
        This registers a given directory for removal on the shutdown of Ganga
        Args:
            shareddir (str): This is the directory which we intend to remove recursively when ganga shuts down
        """
        if shareddir not in self.removal_list:
            self.removal_list.append(shareddir)

    @synchronised
    def cleanUpOrphans(self, orphans=None):
        """
        This cleans up the orphan share dir objects on shutdown
        Args:
            orphans (list): An optional list of the orphans to remove from the list
        """

        if orphans:
            to_remove = orphans
        else:
            to_remove = self.removal_list

        for shareddir in to_remove:
            if os.path.exists(os.path.join(getSharedPath(), shareddir)):
                try:
                    shutil.rmtree(os.path.join(getSharedPath(), shareddir))
                except:
                    logger.error("Failed to remove Orphaned Shared Dir: %s" % shareddir)

        for shareddir in to_remove:
            if shareddir in self.removal_list:
                self.removal_list.remove(shareddir)

    @synchronised
    def counterVal(self, shareddir):
        """Return the current counter value for the named shareddir"""
        from GangaCore.GPIDev.Lib.File import getSharedPath
        shareddir = os.path.join(getSharedPath(), os.path.basename(shareddir.name))
        basedir = os.path.basename(shareddir.name)
        return self.__getName()[basedir]

    @synchronised
    def increase(self, shareddir, force=False):
        """Increase the reference counter for a given shared directory by 1. If the directory
        doesn't currently have a reference counter, one is initialised with a value of 1.
        The shareddir must exist for a reference counter to be initialised (unless Force=True).
        Sharedir should be given relative to the user's shared directory repository, which can
        be discovered by calling 'shareref'.
        Args:
            shareddir (ShareDir): This is the shareddir which we're registering
            force (bool): Ignore whether the directory exists on disk or not
        """
        logger.debug("running increase() in prepregistry")
        self._getSessionLock()


        from GangaCore.GPIDev.Lib.File import getSharedPath
        shareddirname = os.path.join(getSharedPath(), os.path.basename(shareddir.name))
        basedir = os.path.basename(shareddirname)
        if os.path.isdir(shareddirname) and force is False:
            if basedir not in self.__getName():
                logger.debug('%s is not stored in the shareref metadata object...adding.' % basedir)
                self.__getName()[basedir] = 1
            else:
                self.__getName()[basedir] += 1
        elif not os.path.isdir(shareddirname) and force == True and basedir != '':
            if basedir not in self.__getName():
                logger.debug('%s is not stored in the shareref metadata object...adding.' % basedir)
                self.__getName()[basedir] = 1
            else:
                self.__getName()[basedir] += 1
        else:
            logger.error('Directory %s does not exist' % shareddirname)

        self._setDirty()
        self._releaseSessionLockAndFlush()

    @synchronised
    def decrease(self, shareddir, remove=0):
        """Reduce the reference counter for a given shared directory by 1. If the current value
        of the counter is 0, the shared object will be removed from the metadata, and the files within
        the shared object directory deleted when Ganga exits. If the optional remove parameter is specified
        the shared directory is removed from the table.
        Args:
            shareddir (ShareDir): This is the shared directory object to reduce the counter for
            remove (int): Effectively used as a bool. Should the directory be removed when the count reaches 0
        """
        self._getSessionLock()

        from GangaCore.GPIDev.Lib.File import getSharedPath
        shareddirname = os.path.join(getSharedPath(), os.path.basename(shareddir.name))
        basedir = os.path.basename(shareddirname)
        # if remove==1, we force the shareref counter to 0
        try:
            if self.__getName()[basedir] > 0:
                if remove == 1:
                    self.__getName()[basedir] = 0
                else:
                    self.__getName()[basedir] -= 1

                if self.__getName()[basedir] == 0:
#                    shutil.rmtree(shareddir, ignore_errors=True)
                    shareddir.remove()
                    logger.info("Removed: %s" % shareddir.name)
        # if we try to decrease a shareref that doesn't exist, we just set the
        # corresponding shareref to 0
        except KeyError as err:
            logger.debug("KeyError: %s" % err)
            self.__getName()[basedir] = 0
            self.cleanUpOrphans([basedir,])

        self._setDirty()
        self._releaseSessionLockAndFlush()

    def lookup(self, sharedir, unprepare=False):
        """
        Report Job, Box and Task repository items which reference a given ShareDir object. 
        The optional parameter 'unprepare=True' can be set to call the unprepare method 
        on the returned objects.
        """
        from GangaCore.Core.GangaRepository import getRegistryProxy
        objectlist = []
        for thing in getRegistryProxy('jobs').select():
            objectlist.append({thing: 'job'})
        for thing in getRegistryProxy('box').select():
            objectlist.append({thing: 'box'})
        for thing in getRegistryProxy('tasks').select():
            objectlist.append({thing: 'task'})

        run_unp = None
        master_index = 0
        for item in objectlist:
            try:
                list(item.keys())[0].is_prepared.name
                if list(item.keys())[0].is_prepared.name == sharedir:
                    logger.info('ShareDir %s is referenced by item #%s in %s repository' % (sharedir, stripProxy(list(item.keys())[0])._registry_id, list(item.values())[0]))
                    run_unp = list(item.keys())[0]
                    master_index += 1

            except AttributeError as err:
                logger.debug("Err: %s" % err)
                try:
                    list(item.keys())[0].application.is_prepared.name
                    if list(item.keys())[0].application.is_prepared.name == sharedir:
                        logger.info('ShareDir %s is referenced by item #%s in %s repository' % (sharedir, stripProxy(list(item.keys())[0])._registry_id, list(item.values())[0]))
                        run_unp = list(item.keys())[0].application
                        master_index += 1
                except AttributeError as err2:
                    logger.debug("Err2: %s" % err2)
                    try:
                        list(item.keys())[0].analysis.application.is_prepared.name
                        if list(item.keys())[0].analysis.application.is_prepared.name == sharedir:
                            logger.info('ShareDir %s is referenced by item #%s in %s repository' % (sharedir, stripProxy(list(item.keys())[0])._registry_id, list(item.values())[0]))
                            run_unp = list(item.keys())[0].analysis.application
                            master_index += 1
                    except AttributeError as err3:
                        logger.debug("Err3: %s" % err3)
                        pass

            if run_unp is not None and unprepare is True:
                logger.info('Unpreparing %s repository object #%s associated with ShareDir %s' % (list(item.values())[0],  stripProxy(list(item.keys())[0])._registry_id, sharedir))
#                stripProxy(item.keys()[0]).unprepare()
                run_unp.unprepare()
                run_unp = None

        if unprepare is not True:
            logger.info('%s item(s) found referencing ShareDir %s', master_index, sharedir)

    @staticmethod
    def to_relative(this_object):
        from GangaCore.GPIDev.Lib.File import getSharedPath
        logger.info('Absolute ShareDir().name attribute found in Job #%s', this_object.id)
        logger.info('Converting to relative path and moving associated directory if it exists.')
        try:
            shutil.move(this_object.is_prepared.name, os.path.join(getSharedPath(), os.path.basename(this_object.is_prepared.name)))
        except OSError as err:
            GangaCore.Utility.logging.log_unknown_exception()
            logger.warn('Unable to move directory %s to %s', this_object.is_prepared.name, os.path.join(getSharedPath(), os.path.basename(this_object.is_prepared.name)))

        try:
            stripProxy(this_object).is_prepared.name = os.path.basename(this_object.is_prepared.name)
        except Exception as err:
            logger.debug("rebuild Error: %s" % err)
            GangaCore.Utility.logging.log_unknown_exception()
            logger.warn("Unable to convert object's is_prepared.name attribute to a relative path")

    def helper(self, this_object, unp=True, numsubjobs=0):
        from GangaCore.GPIDev.Lib.File import getSharedPath
        shareddir = os.path.join(getSharedPath(), os.path.basename(this_object))
        logger.debug('Adding %s to the shareref table.' % shareddir)
        if os.path.basename(this_object) in self.__getName():
            self.__getName()[os.path.basename(this_object)] += 1
        else:
            self.__getName()[os.path.basename(this_object)] = 1
        if numsubjobs > 0:
            self.__getName()[os.path.basename(this_object)] += numsubjobs
        if not os.path.isdir(shareddir) and os.path.basename(this_object) not in lookup_input:
           logger.info('Shared directory %s not found on disk.' % shareddir)
           if unp == True:
               lookup_input.append(os.path.basename(this_object))

    def rebuild(self, unprepare=True, rmdir=False):
        """Rebuild the shareref table. 
        Clears the shareref table and then rebuilds it by iterating over all Ganga Objects 
        in the Job, Box and Task repositories. If an object has a ShareDir associated with it, 
        that ShareDir is added into the shareref table (the reference counter being incremented 
        accordingly). If called with the optional parameter 'unprepare=False', objects whose
        ShareDirs are not present on disk will not be unprepared. Note that the default behaviour
        is unprepare=True, i.e. the job/application would be unprepared.
        After all Job/Box/Task objects have been checked, the inverse operation is performed, 
        i.e., for each directory in the ShareDir repository, a check is done to ensure there 
        is a matching entry in the shareref table. If not, and the optional parameter 
        'rmdir=True' is set, then the (orphaned) ShareDir will removed from the filesystem. 
        Otherwise, it will be added to the shareref table with a reference count of zero; 
        this results in the directory being deleted upon Ganga exit.
        """
        self._getSessionLock()
        # clear the shareref table
        self.name = {}
        lookup_input = []

        from GangaCore.GPIDev.Lib.File import getSharedPath
        from GangaCore.Core.GangaRepository import getRegistryProxy

        objectlist = []
        for thing in getRegistryProxy('jobs').select():
            objectlist.append({thing: 'job'})
        for thing in getRegistryProxy('box').select():
            objectlist.append({thing: 'box'})
        for thing in getRegistryProxy('tasks').select():
            objectlist.append({thing: 'task'})

        for item in objectlist:
            shortname = None
            try:
                shortname = list(item.keys())[0].is_prepared.name
            except AttributeError as err:
                logger.debug("Err: %s" % err)
                try:
                    shortname = list(item.keys())[0].application.is_prepared.name
                except AttributeError as err2:
                    logger.debug("Err2: %s" % err2)
                    try:
                        shortname = list(item.keys())[0].analysis.application.is_prepared.name
                    except AttributeError as err3:
                        logger.debug("Err3: %s" % err3)
                        pass
            try:
                if shortname is not None and shortname is not True:
                    if os.path.basename(shortname) != shortname:
                        self.to_relative(list(item.keys())[0].is_prepared)
                    try:
                        numsubjobs = len(list(item.keys())[0].subjobs.ids())
                    except Exception as err:
                        logger.debug("Path Error: %s" % err)
                        GangaCore.Utility.logging.log_unknown_exception()
                        numsubjobs = 0
                    self.helper(shortname, unp=unprepare, numsubjobs=numsubjobs)
            except Exception as err:
                logger.debug("-Error: %s" % err)
                GangaCore.Utility.logging.log_unknown_exception()
                pass

        # here we iterate over the lookup_input list and unprepare as
        # necessary.
        for item in lookup_input:
            logger.info('Unpreparing objects referencing ShareDir %s' % item)
            self.lookup(sharedir=item, unprepare=True)

        # check to see that all sharedirs have an entry in the shareref. Otherwise, set their ref counter to 0
        # so the user is made aware of them at shutdown
        for this_dir in os.listdir(getSharedPath()):
            if this_dir not in list(self.__getName().keys()) and rmdir is False:
                logger.debug("%s isn't referenced by a GangaObject in the Job or Box repository." % this_dir)
                self.__getName()[this_dir] = 0
            elif this_dir not in self.__getName() and rmdir is True:
                logger.debug("%s isn't referenced by a GangaObject in the Job or Box repository. Removing directory." % this_dir)
                shutil.rmtree(os.path.join(getSharedPath(), this_dir))

        self._setDirty()
        self._releaseSessionLockAndFlush()


    @staticmethod
    def yes_no(question, default='none', shareddir=''):
        """Check whether the user wants to delete sharedirs which are no longer referenced by any Ganga object"""
        valid = {"yes": "yes", "y": "yes", "no": "no",
                     "n": "no", "none": "none", "all": "all"}
        if default == 'none':
            prompt = '(Yes/No/All/[NONE])'
        elif default == 'yes':
            prompt = '([YES]/No/All/None)'
        elif default == 'no':
            prompt = '(Yes/[NO]/All/None)'
        else:
            raise ValueError("Invalid default answer: '%s'" % default)
        while True:
            logger.info('%s no longer being referenced by any objects. Delete directory?' % shareddir)
            logger.info(question + prompt)
            answer = input().lower()
            if answer == '':
                return default
            elif answer in valid:
                return valid[answer]
            else:
                logger.warn("Please respond with 'Yes/y', 'No/n', 'All' or 'None'")

    def closedown(self):
        """Cleans up the Shared Directory registry upon shutdown of the registry, ie. when exiting a Ganga session."""

        #stripProxy(self)._getRegistry()._hasStarted = True
        from GangaCore.GPIDev.Lib.File import getSharedPath

        delete_share_config = GangaCore.Utility.Config.getConfig('Configuration')['deleteUnusedShareDir']
        if delete_share_config == 'ask':
            ask_delete = 'Ask'
            default = 'none'
        elif delete_share_config == 'never':
            ask_delete = 'none'
        elif delete_share_config == 'always':
            ask_delete = 'all'
        else:
            ask_delete = 'Ask'
            default = 'none'

        # list of keys to be removed from the shareref table
        cleanup_list = []

        try:
            ## FIXME. this triggers maximum recusion depth bug on shutdown in some situations! rcurrie
            all_dirs = copy.deepcopy(list(self.__getName().keys()))
        except:
            all_dirs = {}
        for shareddir in all_dirs:
            full_shareddir_path = os.path.join(getSharedPath(), shareddir)
            # for each sharedir in the shareref table that also exists in the
            # filesystem
            if self.__getName()[shareddir] == 0 and os.path.isdir(full_shareddir_path):
                if ask_delete == 'Ask':
                    ask_delete = self.yes_no('', default=default, shareddir=shareddir)
                if ask_delete == 'yes':
                    shutil.rmtree(full_shareddir_path)
                    if shareddir not in cleanup_list:
                        cleanup_list.append(shareddir)
                    ask_delete = 'Ask'
                    default = 'yes'
                    logger.debug('Deleting Sharedir %s because it is not referenced by a persisted Ganga object', shareddir)

                if ask_delete == 'no':
                    ask_delete = 'Ask'
                    default = 'no'
                    logger.debug('Keeping ShareDir %s even though it is not referenced by a persisted Ganga object', shareddir)

                if ask_delete == 'all':
                    shutil.rmtree(full_shareddir_path)
                    if shareddir not in cleanup_list:
                        cleanup_list.append(shareddir)
                    logger.debug('Deleting Sharedir %s because it is not referenced by a persisted Ganga object', shareddir)

                if ask_delete == 'none':
                    default = 'none'
                    logger.debug('Keeping ShareDir %s even though it is not referenced by a persisted Ganga object', shareddir)

            ## DISABLED BY RCURRIE
            # if the sharedir in the table doesn't exist on the filesytem, and the reference counter is > 0,
            # we need to unprepare any associated jobs
            #logger.debug("Examining: %s" % full_shareddir_path)
            #logger.debug("shareddir: %s" % shareddir)
            #logger.debug("cleanup_list: %s" % cleanup_list)
            if not os.path.isdir(full_shareddir_path) and shareddir not in cleanup_list:
            #    logger.info('%s not found on disk. Removing entry from shareref table and unpreparing any associated Ganga objects.' % shareddir)
            #    self.lookup(sharedir=shareddir, unprepare=True)
                cleanup_list.append(shareddir)
            ## DISABLED BY RCURRIE

        self._getSessionLock()
        for element in cleanup_list:
            del self.name[element]
        allnames = copy.deepcopy(self.__getName())
        for element in allnames:
            del self.name[element]
        self._setDirty()
        #self._releaseWriteAccess()

        #stripProxy(self)._getRegistry()._hasStarted = False

    def ls(self, shareddir, print_files=True):
        """
        Print the contents of the given shared directory, which can be specified as relative to the shared
        directories repository, or absolute.
        """
        shareddir_root = shareddir
        full_shareddir_path = os.path.join(getSharedPath(), os.path.basename(shareddir))

        if os.path.isdir(full_shareddir_path):
            cmd = "find '%s'" % (full_shareddir_path)
            files = os.popen(cmd).read().strip().split('\n')
            padding = '|  '
            for file in files:
                level = file.count(os.sep)
                level = level - 6
                pieces = file.split(os.sep)
                symbol = {0: '', 1: '/'}[os.path.isdir(file)]
                if not print_files and symbol != '/':
                    continue
                logger.info(padding * level + pieces[-1] + symbol)

    def _display(self, interactive=0):
        """Prints content of the shareref metadata in a well formatted way.
        """

        if len(self.__getName()) > 0:
            from GangaCore.GPIDev.Lib.File import getSharedPath
            fstring = " %48s | %20s |  %15s"
            disp_string = fstring % (
                "Shared directory", "Date created", "Reference count\n")
#           print fstring % (" ", " "," ")
            disp_string += fstring % ("------------------------------------------------",
                                      "--------------------", "---------------\n")
            zero_ref = False
            unsorted = []
            all_elements = copy.deepcopy(self.__getName())
            for element in all_elements:
                full_shareddir_path = os.path.join(getSharedPath(), os.path.basename(element))
                if os.path.isdir(full_shareddir_path):
                    unsorted.append(shareref_data(os.path.basename(element), int(
                        os.path.getctime(full_shareddir_path)), self.__getName()[element]))
                else:
                    unsorted.append(shareref_data(os.path.basename(element), "Directory not found", self.__getName()[element]))
            decorated = sorted((name.date, i, name) for i, name in enumerate(unsorted))
            sorted_refs = [name for date, i, name in decorated]
            for line in sorted_refs:
                if isinstance(line.date, int):
                    tmp_string = fstring % (os.path.basename(line.name),
                                            time.strftime(
                                                "%d %b %Y %H:%M:%S", time.localtime(line.date)),
                                            line.counter)
                else:
                    tmp_string = fstring % (os.path.basename(line.name),
                                            line.date,
                                            line.counter)

                if (line.counter == 0) or (isinstance(line.date, str)):
                    from GangaCore.Utility.ColourText import ANSIMarkup, NoMarkup, Foreground, Background, Effects
                    fg = Foreground()
                    fg = Background()
                    fx = Effects()
                    if interactive:
                        m = ANSIMarkup()
                    else:
                        m = NoMarkup()
                    disp_string += fg.red + tmp_string + fx.normal + '\n'
                    #disp_string += m(tmp_string,code=fg.red)
                    if (line.counter == 0):
                        zero_ref = True
                else:
                    disp_string += tmp_string + '\n'

            disp_string += "\nThe shared directory repository is rooted at " + \
                getSharedPath() + "\n"
            if zero_ref:
                disp_string += "\nShared directories with a zero reference count will be removed when Ganga exits.\n"
        else:
            disp_string = "No objects stored in the shared directory."

        return disp_string

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text('prep registry...')
            return
        p.text(self._display(interactive=True))

    # rcurrie Adding this due to strange bug but assuming it should be false
    # due to setRegistry(None)
    def _registry_locked(self):
        return False

    def _proxy_display(self, interactive=1):
        return self._display(interactive=interactive)


class shareref_data(object):

    def __init__(self, name, date, counter):
        self.name = name
        self.date = date
        self.counter = counter

    def __repr__(self):
        return repr((self.name, self.date, self.counter))


class _proxy_display(object):

    def __get__(self, obj, cls):
        if obj is None:
            return stripProxy(cls)._proxy_display
        return stripProxy(obj)._proxy_display

ShareRef.__str__ = ShareRef._display




from GangaCore.Utility.logging import getLogger
logger = getLogger(modulename=True)

class TestMonitoringService:
    def __init__(self, jobID, workdir='',application=''):
        self._jobID = jobID
        self.workdir = workdir
        logger.info('TestMonitoringService initialized: jobID=%s workdir=%s application=%s'%(jobID,workdir,application))
        
    def send(self):
        logger.info('TestMonitoringService.send')

    def event(self,what='',values={}):
        logger.info('TestMonitoringService.event: what=%s values=%s'%(str(what),str(values)))

    def sandBoxModule(self):
        import GangaTest.Lib.TestMonitoringService
        return [
               GangaTest,
               GangaTest.Lib,
               GangaTest.Lib.TestMonitoringService,
               GangaTest.Lib.TestMonitoringService.TestMonitoringService
               ]
        





#Bail out when loading this module as it is not python3 compliant
from GangaCore.Core.exceptions import PluginError
raise PluginError("The GangaPanda module has not been upgraded for python 3. The last python 2 Ganga version is 7.1.15 . Please contact the ganga devs to discuss updating this module.")

## Config options
from GangaCore.Utility.Config import makeConfig
from GangaCore.Utility.logging import getLogger
logger = getLogger()

# -------------------------------------------------
# Panda Options
config = makeConfig('Panda','Panda backend configuration parameters')
config.addOption( 'prodSourceLabelBuild', 'panda', 'prodSourceLabelBuild')
config.addOption( 'prodSourceLabelRun', 'user', 'prodSourceLabelRun')
config.addOption( 'assignedPriorityBuild', 2000, 'assignedPriorityBuild' )
config.addOption( 'assignedPriorityRun', 1000, 'assignedPriorityRun' )
config.addOption( 'processingType', 'ganga', 'processingType' )
config.addOption( 'specialHandling', '', 'specialHandling - Expert only.' )
config.addOption( 'enableDownloadLogs', False , 'enableDownloadLogs' )
config.addOption( 'trustIS', True , 'Trust the Information System' )
config.addOption( 'serverMaxJobs', 5000 , 'Maximum number of subjobs to send to the Panda server' )
config.addOption( 'chirpconfig', '' , 'Configuration string for chirp data output, e.g. "chirp^etpgrid01.garching.physik.uni-muenchen.de^/tanyasandoval^-d chirp" ' )
config.addOption( 'chirpserver', '' , 'Configuration string for the chirp server, e.g. "voatlas92.cern.ch". If this variable is set config.Panda.chirpconfig is filled and chirp output will be enabled.' )
config.addOption( 'siteType', 'analysis' , 'Expert only.' )
config.addOption( 'baseURL', '' , 'Expert only.' )
config.addOption( 'baseURLSSL', '' , 'Expert only.' )
config.addOption( 'AllowDirectSubmission', False, 'Deprecated. Please use the Jedi backend' )
config.addOption('AGISJSONFile', '/tmp/agis_pandaresources.json', 'Local AGIS JSON file for queue -> site mapping')

# -------------------------------------------------
# Jedi Options
config = makeConfig('Jedi','Jedi backend configuration parameters')
config.addOption( 'prodSourceLabelBuild', 'panda', 'prodSourceLabelBuild')
config.addOption( 'prodSourceLabelRun', 'user', 'prodSourceLabelRun')
config.addOption( 'assignedPriorityBuild', 2000, 'assignedPriorityBuild' )
config.addOption( 'assignedPriorityRun', 1000, 'assignedPriorityRun' )
config.addOption( 'processingType', 'ganga', 'processingType' )
config.addOption( 'enableDownloadLogs', False , 'enableDownloadLogs' )
config.addOption( 'trustIS', True , 'Trust the Information System' )
config.addOption( 'serverMaxJobs', 5000 , 'Maximum number of subjobs to send to the Panda server' )
config.addOption( 'chirpconfig', '' , 'Configuration string for chirp data output, e.g. "chirp^etpgrid01.garching.physik.uni-muenchen.de^/tanyasandoval^-d chirp" ' )
config.addOption( 'chirpserver', '' , 'Configuration string for the chirp server, e.g. "voatlas92.cern.ch". If this variable is set config.Panda.chirpconfig is filled and chirp output will be enabled.' )
config.addOption( 'siteType', 'analysis' , 'Expert only.' )



def standardSetup():
    import PACKAGE
    PACKAGE.standardSetup()

    
def loadPlugins(c):

    import sys
    from GangaCore.Utility.logging import getLogger

    try:
        import Lib.Panda
        import Lib.Jedi
    except SystemExit:
        from GangaCore.Core.exceptions import ApplicationConfigurationError
        import commands
        (s,o) = commands.getstatusoutput('curl --version')
        if (s):
            raise ApplicationConfigurationError("Couldn't load Panda Client: ensure 'curl' is available")
        else:
            raise ApplicationConfigurationError("Couldn't load Panda Client")
    import Lib.Athena
    import Lib.Executable
    import Lib.ProdTrans

    return None




##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: ICredential.py,v 1.6 2009/05/20 13:40:22 moscicki Exp $
##########################################################################
#
# File: ICredential.py
# Author: K. Harrison
# Created: 060607
#
# 06/07/2006 KH:  Changed to GangaCore.Utility.Shell for shell commands
#                 Redefined create and renew methods
#
# 07/08/2006 KH:  Added isValid() method
#
# 08/08/2006 KH:  Corrected errors in methods isValid() and timeInSeconds()
#                 => thanks to HCL and CLT
#                 Make shell an ICredential property
#
# 06/09/2006 KH:  Changes to allow minimum validity of credential
#                 and validity at creation to be specified independently
#                 Changes for monitoring component from CLT
#
# 25/09/2006 KH:  Changed method isValid(), so that default validity is
#                 value of self.minValidity
#
# 10/10/2006 KH:  Made changes to warnings for credential renewal
#
# 23/11/2006 KH:  Added "pipe" keyword to option dictionary of ICommandSet
#                 Added method to determine if credential is available
#                 with system/configuration used
#                 (requests from CLT)
#
# 28/02/2007 CLT: Modified GUI-specific and CLI-specific routines in create(),
#                 to reflect the use of currentOpts dictionary when building
#                 command options
#                 GUI-Specific code in create() updated to make use of the new
#                 Credentials Manager panel in the GUI
#
# 27/07/2007 AM:  Updated the renew() method to notify the
#                 InternalServices.Coordinator when credentials are
#                 detected by the monitoring loop as being invalid
#                 so the internal services to be disabled
#
# 07/12/2007 KH:  Made ICommandSet a component class
#
# 17/12/2007 KH:  Added function registerCommandSet for registering plugins
#                 of category "credential_commands"
#
# 13/03/2008 KH:  Update for change in configuration system
#                 (use "defaults_" instead of "_Properties")
#
# 18/03/2009 MWS: Added the 'log' option to isValid and shifted the expire
#                 message here
#
# 15/10/2009 MWS: Added possibility to force new check of credentials,
#                 rather than relying on cache

"""Module defining interface class for working with credentials"""

__author__ = "K.Harrison <Harrison@hep.phy.cam.ac.uk>"
__date__ = "05 November 2009"
__version__ = "1.11"

import os
import threading
import time
import tempfile

from GangaCore.GPIDev.Base.Objects import GangaObject
from GangaCore.GPIDev.Schema import ComponentItem, Schema, SimpleItem, Version
from GangaCore.Utility.Config import ConfigError, getConfig
from GangaCore.Utility.logging import getLogger
from GangaCore.Utility.Plugin.GangaPlugin import allPlugins
from GangaCore.Utility.Shell import Shell

from GangaCore.Core.InternalServices import Coordinator

logger = getLogger()
logTimeStamp = 0.0
logRepeatDuration = 120.0


def registerCommandSet(commandClass=None):

    try:
        pluginLoaded = allPlugins.all_dict\
            ["credential_commands"][commandClass._name]
    except KeyError:
        allPlugins.add(commandClass, "credential_commands", commandClass._name)

    return None


class ICommandSet(GangaObject):

    """
    Class used to define shell commands and options for working with credentials
    """
    _schema = Schema(Version(1, 0), {
        "init": SimpleItem(defvalue="",
                           doc="Command for creating/initialising credential"),
        "info": SimpleItem(defvalue="",
                           doc="Command for obtaining information about credential"),
        "destroy": SimpleItem(defvalue="",
                              doc="Command for destroying credential"),
        "init_parameters": SimpleItem(defvalue={},
                                      doc="Dictionary of parameter-value pairs to pass to init command"),
        "destroy_parameters": SimpleItem(defvalue={},
                                         doc="Dictionary of parameter-value pairs to pass to destroy command"),
        "info_parameters": SimpleItem(defvalue={},
                                      doc="Dictionary mapping from Ganga credential properties to command-line options"),
    })

    _category = "credential_commands"
    _name = "ICommandSet"
    _hidden = 1
    _enable_config = 1

    def __init__(self):
        super(ICommandSet, self).__init__()
        return

    def setConfigValues(self):
        """
        Update properties using values from relevant section of config file.
        """
        section = "defaults_%s" % self._name
        config = getConfig(section)
        for attribute in self._schema.datadict.keys():
            try:
                value = config[attribute]
                try:
                    value = eval(value)
                except:
                    pass
                setattr(self, attribute, value)
            except ConfigError:
                pass

registerCommandSet(ICommandSet)


class ICredential(GangaObject):

    """
    Interface class for working with credentials
    """

    _schema = Schema(Version(1, 0), {
        "maxTry": SimpleItem(defvalue=1, typelist=[int],
                             doc="Number of password attempts allowed when creating credential"),
        "minValidity": SimpleItem(defvalue="00:15", typelist=[str],
                                  doc="Default minimum validity"),
        "validityAtCreation": SimpleItem(defvalue="24:00", typelist=[str],
                                         doc="Default credential validity at creation"),
        "command": ComponentItem(category="credential_commands",
                                 defvalue="ICommandSet",
                                 doc="Set of commands to be used for credential-related operations")
    })

    _category = "credentials"
    _name = "ICredential"
    _hidden = 1

    _exportmethods = ["create", "destroy", "isAvailable", "isValid",
                      "location", "renew", "timeleft"]

    def __init__(self):
        super(ICredential, self).__init__()
        self.shell = Shell()
        self.inputPW_Widget = None
        return

    def create(self, validity="", maxTry=0, minValidity="", check=False):
        """
        Create credential.

        Arguments other than self:
           validity    - Validity with which credential should be created,
                         specified as string of format "hh:mm"
                         [ Defaults to value of self.validityAtCreation ]
           maxTry      - Number of password attempts allowed
                         [ Defaults to value of self.maxTry ]
           minValidity - Minimum validity in case checking of
                         pre-existing credential is performed,
                         specified as strong of format "hh:mm"
                         [ Defaults to value of self.minValidity ]
           check       - Flag to request checking of pre-existing
                         credential; if flag is set to true, then new
                         credential is created only if the validity of
                         any pre-existing credential is less than the
                         value of minValidity
                         [ Default: False ] 

        Note: create is the same as renew, except for the default value of check

        Return value: True if credential is created successfully, and False
        otherwise.
        """
        global logTimeStamp

        dummy = False
        if not self.command.init:
            dummy = True
        if "valid" in self.command.init_parameters:
            if not self.command.init_parameters["valid"]:
                dummy = True

        if dummy:
            logger.warning("Dummy CommandSet used - no credential created")
            return False

        if not maxTry:
            maxTry = self.maxTry

        if not minValidity:
            minValidity = self.minValidity

        if not validity:
            validity = self.validityAtCreation

        validityInSeconds = self.timeInSeconds(validity)

        if not validityInSeconds:
            logger.warning("Problems with requested validity: %s" % str(validity))
            return False
        if check and self.isValid(minValidity):
            return True

        ntry = 0

        while ntry < maxTry:

            ntry = ntry + 1
            # Test if GUI widget is to be used.
            if self.inputPW_Widget:
                # Since self.inputPW_Widget is called, current arguments are
                # ignored since renew() and create() in GUI mode will not be
                # called with any arguments.
                #proxy_obj = self._proxyObject ## This is removed to get rid of ref to _proxyObject
                proxy_obj = self
                if self.inputPW_Widget.ask(proxy_obj):
                    logger.dg("Proceeding to retrieve password from inputPW_Widget.")
                    __pw = self.inputPW_Widget.getPassword(proxy_obj)
                    if not __pw:
                        logger.warning("Password/passphrase expected!")
                        return False
                    try:
                        tFile = tempfile.NamedTemporaryFile()
                        tFile.write(__pw)
                        tFile.flush()
                    except:
                        del __pw
                        logger.warning("Could not create secure temporary file for password!")
                        return False
                    del __pw
                else:
                    # Current credential modification denied for various reasons.
                    # see GangaGUI.customDialogs.ask() method for more details.
                    return False
                # self.inputPW_Widget.ask() may have modified parameters.
                # Calling buildOpts() to take them into account.
                self.buildOpts(self.command.init, False)
                # Create initialisation list with the 'pipe' parameter
                initList = [self.command.init, self.command.init_parameters["pipe"]]
                # Append option value pairs
                for optName, optVal in self.command.currentOpts.items():
                    initList.append("%s %s" % (optName, optVal))
                status = self.shell.system("cat %s|%s" % (tFile.name, " ".join(initList)))
                tFile.close()
                # self.inputPW_Widget dialog postprocessing.
                # E.g. disable autorenew mechanism if status != 0.
                self.inputPW_Widget.renewalStatus(proxy_obj, status)
                if status == 0:
                    logger.info("%s creation/renewal successful." % self._name)
                    return True
                else:
                    logger.warning("%s creation/renewal failed [%s]." % (self._name, status))
                    return False
            else:  # Non-GUI credential renewal/creation
                # Check if renewal is from main process (i.e. by bootstrap or
                # user)
                if threading.currentThread().getName() == 'MainThread' or\
                        threading.currentThread().getName().startswith('GANGA_Update_Thread_Ganga_Worker_'):
                    if "valid" in self.command.init_parameters:
                        self.command.currentOpts[self.command.init_parameters['valid']] = validity
                    initList = [self.command.init]
                    # Append option value pairs
                    for optName, optVal in self.command.currentOpts.items():
                        initList.append("%s %s" % (optName, optVal))
                    status = self.shell.system(" ".join(initList))
                    if status == 0:
                        logger.info("%s creation/renewal successful." % self._name)
                        return True
                    else:
                        logger.warning("%s creation/renewal failed [%s]." % (self._name, status))
                # create initiated from worker thread from monitoring
                # component.
                else:
                    currTime = time.time()
                    if currTime - logTimeStamp >= logRepeatDuration:
                        logTimeStamp = currTime

                        # Check validity but print logging messages this time
                        self.isValid("", True)
                        _credentialObject = self._name[0].lower() + self._name[1:]
                        logger.warning("Renew by typing '%s.renew()' at the prompt." % (_credentialObject))

                        # notify the Core that the credential is not valid
                        _validity = self.timeInSeconds(self.timeleft())
                        _minValidity = self.timeInSeconds(minValidity) / 2.
                        if _validity <= max(120, _minValidity):
                            Coordinator.notifyInvalidCredential(self)

                    return True

        logger.warning("%s creation/renewal attempts exceeded %s tries!" % (self._name, maxTry))
        return False

    def destroy(self, allowed_exit=[0]):
        """
        Destroy credential

        Argument other than self:
           allowed_exit - List of exit codes accepted without error
                          when issuing system command for destroying credential

        Return value: False if command for destroying credential is undefined,
                      or True otherwise
        """

        if not self.command.destroy:
            logger.warning("Dummy CommandSet used - no credential created")
            return False

        destroyList = [self.command.destroy]
        for optName, optVal in self.command.destroyOpts.items():
            destroyList.append("%s %s" % (optName, optVal))

        Coordinator.notifyInvalidCredential(self)

        status, output, message = \
            self.shell.cmd1(" ".join(destroyList), allowed_exit)
        proxyPath = self.location()
        if proxyPath:
            os.remove(proxyPath)
        return True

    def isAvailable(self):
        """
        Check whether credential is available with system/configuration used

        No arguments other than self

        Return value: True if credential is available, false otherwise
        """

        logger.warning("Dummy method used - this always returns True")

        return True

    def isValid(self, validity="", log=False, force_check=False):
        """
        Check validity

        Arguments other than self:
           validity    - Minimum time for which credential should be valid,
                         specified as string of format "hh:mm"
                         [ Defaults to valud of self.minValidity ]

           log         - Print logger messages if credential not valid 

           force_check - Force credential check, rather than relying on cache

        Return value: True if credential is valid for required time, False
        otherwise.
        """

        valid = True

        if not validity or validity is None:
            validity = self.minValidity
        validityInSeconds = self.timeInSeconds(validity)
        timeleft = self.timeleft(force_check=force_check)

        if not timeleft:
            valid = False
        else:
            timeleftInSeconds = self.timeInSeconds(timeleft)
            if timeleftInSeconds <= validityInSeconds:
                valid = False

        if not valid and log:
            _tl = self.timeleft(force_check=force_check)
            if _tl == "-1" or _tl == "0:00:00":
                _expiryStatement = "has expired!"
            else:
                _expiryStatement = "will expire in %s!" % _tl

            itemList = []
            text = self._name[0]
            for i in range(len(self._name) - 1):
                character = self._name[i + 1]
                if character.isupper():
                    itemList.append(text)
                    text = character.lower()
                else:
                    text = "".join([text, character])
            itemList.append(text)
            _credentialName = " ".join(itemList)

            logger.warning("%s %s" %
                           (_credentialName, _expiryStatement))

        return valid

    def location(self):
        """
        Determine credential location

        No arguments other than self

        Return value: Path to credential if found, or empty string otherwise
        """

        return ""

    def renew(self, validity="", maxTry=0, minValidity="",
              check=True):
        """
        Renew credential.

        Arguments other than self:
           validity    - Validity with which credential should be created,
                         specified as string of format "hh:mm"
                         [ Defaults to value of self.validityAtCreation ]
           maxTry      - Number of password attempts allowed
                         [ Defaults to value of self.maxTry ]
           minValidity - Minimum validity in case checking of
                         pre-existing credential is performed,
                         specified as strong of format "hh:mm"
                         [ Defaults to value of self.minValidity ]
           check       - Flag to request checking of pre-existing
                         credential; if flag is set to true, then new
                         credential is created only if the validity of
                         any pre-existing credential is less than the
                         value of minValidity
                         [ Default: True ] 

        Note: renew is the same as create, except for the default value of check

        Return value: True if new credential is created successfully, and False
        otherwise.
        """
        status = self.create(validity, maxTry, minValidity, check)

        return status

    def timeInSeconds(self, timeString=""):
        """
        Convert time string to time in seconds

        Arguments other than self:
           timeString - Time specified as string of format "hh:mm:ss"

        Return value: Time in seconds (integer)
        """

        totalTime = 0
        timeList = timeString.split(":")
        if len(timeList) >= 1:
            totalTime = totalTime + int(timeList[0]) * 60 * 60
        if len(timeList) >= 2:
            totalTime = totalTime + int(timeList[1]) * 60
        if len(timeList) >= 3:
            totalTime = totalTime + int(timeList[2])

        return totalTime

    def timeleft(self, units="hh:mm:ss", force_check=False):
        """
        Check time for which credential is valid.

        Arguments other than self:
           units       - String specifying units in which time is returned

           force_check - Force credential check, rather than relying on cache

        Allowed values for units are:
           "hours"              - time returned as in hours
           "minutes"            - time returned in minutes
           "seconds"            - time returned in seconds
           "hh:mm:ss" [default] - time returned as hours, minutes seconds


        Return value: Credential validity as string giving time in requested
           units, or empty string if command for querying credential validity
           is unavailable
        """

        timeRemaining = self.timeleftInHMS(force_check=force_check)
        if timeRemaining not in ["", "-1"]:
            if units in ["hours", "minutes", "seconds"]:
                timeleftInSeconds = self.timeInSeconds(timeRemaining)
                if "seconds" == units:
                    timeRemaining = "%.2f" % (timeleftInSeconds)
                elif "minutes" == units:
                    timeRemaining = "%.2f" % (timeleftInSeconds / 60.)
                elif "hours" == units:
                    timeRemaining = "%.2f" % (timeleftInSeconds / (60. * 60.))

        return timeRemaining

    def timeleftInHMS(self, force_check=False):
        """
        Determine remaining validity of credential in hours, minutes and seconds

        Argument other than self:
           force_check - Force credential check, rather than relying on cache

        Return value: String giving credential validity, or empty string
           if command for querying credential validity is unavailable
        """
        logger.warning("Dummy method used - no information returned")
        return ""
